import os

from dynaconf import Dynaconf
from loguru import logger as log


log.remove() # remove default handler to prevent printing to console
log.add('logs/logs.log', level='DEBUG', encoding="utf8")

current_directory = os.path.dirname(os.path.realpath(__file__))
config_path = os.path.abspath(os.path.join(current_directory, 'config'))
settings_file_name = os.path.join(config_path, 'settings.toml')
secrets_file_name = os.path.join(config_path, '.secrets.toml')
plugins_path = os.path.abspath(os.path.join(current_directory, 'plugins'))

if not os.path.isfile(settings_file_name):
    print(f'No config file found at path: {settings_file_name}')
    exit()


def generate_secrets_file():
    with open(secrets_file_name, 'w') as f:
        f.write(
            """
[default]
SPOTIFY_CLIENT_ID = ""
SPOTIFY_CLIENT_SECRET = ""

DEEZER_APP_ID = 0
DEEZER_APP_SECRET = ""
DEEZER_ACCESS_TOKEN = ""           
            """
        )
    pass

if not os.path.isfile(secrets_file_name):
    print(f'No secrets file found at path: {secrets_file_name}')
    generate_secrets_file()
    print(f'New file generated. Please, fill secrets data.\n')


settings = Dynaconf(
    envvar_prefix="SPOTY",
    settings_files=[settings_file_name, secrets_file_name],
)





from spoty.commands.first_list_commands import \
    count_command, \
    delete_command, \
    copy_command, \
    move_command, \
    export_command, \
    import_deezer_command, \
    import_spotify_command, \
    sync_spotify_command, \
    print_command, \
    create_m3u8_command
from spoty.utils import SpotyContext
import click


@click.group("get-unique")
@click.option('--fist-list', '--fl', '-f', is_flag=True,
              help='Get unique tracks only from first collected list.')
@click.option('--second-list', '--sl', '-s', is_flag=True,
              help='Get unique tracks only from second collected list.')
@click.pass_obj
def get_unique(context: SpotyContext,
               fist_list,
               second_list
               ):
    """
Get unique tracks (not duplicated) for further actions (see next commands).
    """

    if not fist_list and not second_list:
        click.echo(f'Please, specify --fist-list or --second-list (or both) for "get-unique" command.', err=True)
        exit()

    context.summary.append("Collecting unique tracks:")

    context.tags_lists.clear()
    context.tags_lists.append([])

    if fist_list:
        context.tags_lists[0].extend(context.unique_first_tracks)
        context.summary.append(f'  {len(context.unique_first_tracks)} unique tracks from first list collected.')
    if second_list:
        context.tags_lists[0].extend(context.unique_second_tracks)
        context.summary.append(f'  {len(context.unique_second_tracks)} unique tracks from second list collected.')

    if len(context.tags_lists[0]) == 0:
        context.summary.append(f'  0 unique tracks collected.')

    context.duplicates_groups = []
    context.unique_first_tracks = []
    context.unique_second_tracks = []

get_unique.add_command(count_command.count_tracks)
get_unique.add_command(print_command.print_tracks)
get_unique.add_command(export_command.export_tracks)
get_unique.add_command(create_m3u8_command.export_tracks)
get_unique.add_command(import_spotify_command.import_spotify)
get_unique.add_command(import_deezer_command.import_deezer)
get_unique.add_command(sync_spotify_command.sync_spotify)
get_unique.add_command(delete_command.delete_tracks)
get_unique.add_command(move_command.move_tracks)
get_unique.add_command(copy_command.copy_tracks)



from spoty import settings
from spoty import log
import spoty.deezer_api
import spoty.csv_playlist
import spoty.deezer_api
import spoty.spotify_api
import spoty.audio_files
import spoty.utils
from spoty.commands import get_group
from spoty.utils import SpotyContext
from spoty.utils import DuplicatesGroup
import click
import os
from datetime import datetime


class TagsToReplace:
    source_tags: dict
    dest_tags: dict
    tags_keys: list

    def __init__(self):
        self.source_tags = {}
        self.dest_tags = {}
        self.tags_keys = []


@click.command("replace-tags")
@click.argument('tags')
@click.option('--confirm', '-y', is_flag=True,
              help='Do not ask for confirmation')
@click.pass_obj
def replace_tags(context: SpotyContext,
                 tags,
                 confirm
                 ):
    """
Get the specified tags from the tracks in the first list and write those tags to the tracks in the second list.

TAGS - Which tags to rewrite (specify separated by commas).
    """
    context.summary.append('Replacing tags:')


    write_tags = tags.split(',')
    for i in range(len(write_tags)):
        write_tags[i] = write_tags[i].upper()

    tags_to_update_list = []
    with click.progressbar(context.duplicates_groups, label='Collecting tags') as bar:
        for group in bar:
            # collect all tags
            for dup_tags in group.def_duplicates:
                if 'SPOTY_FILE_NAME' in dup_tags:  # is local file
                    tags_keys = []
                    for tag in write_tags:
                        if (tag in group.source_tags and tag not in dup_tags) \
                                or (tag not in group.source_tags and tag in dup_tags) \
                                or (tag in group.source_tags and tag in dup_tags and dup_tags[tag] != group.source_tags[
                            tag]):
                            tags_keys.append(tag)
                    if len(tags_keys) > 0:
                        t = TagsToReplace()
                        t.source_tags = group.source_tags
                        t.dest_tags = dup_tags
                        t.tags_keys = tags_keys
                        tags_to_update_list.append(t)

            for dup_tags in group.prob_duplicates:
                if 'SPOTY_FILE_NAME' in dup_tags:  # is local file
                    tags_keys = []
                    for tag in write_tags:
                        if (tag in group.source_tags and tag not in dup_tags) \
                                or (tag not in group.source_tags and tag in dup_tags) \
                                or (tag in group.source_tags and tag in dup_tags and dup_tags[tag] != group.source_tags[
                            tag]):
                            tags_keys.append(tag)
                    if len(tags_keys) > 0:
                        t = TagsToReplace()
                        t.source_tags = group.source_tags
                        t.dest_tags = dup_tags
                        t.tags_keys = tags_keys
                        tags_to_update_list.append(t)

    if len(tags_to_update_list) == 0:
        context.summary.append("  No tags to update found.")
    else:

        click.echo('Next audio files will be edited:')

        for t in tags_to_update_list:
            click.echo(f'  {t.dest_tags["SPOTY_FILE_NAME"]}:')
            for tag in t.tags_keys:
                was = t.dest_tags[tag] if tag in t.dest_tags else ""
                new = t.source_tags[tag] if tag in t.source_tags else ""
                click.echo(f'    {tag}: was: "{was}"')
                click.echo(f'    {tag}: new: "{new}"')

        if not confirm:
            click.confirm(f'Are you sure you want to edit tags in {len(tags_to_update_list)} audio files?', abort=True)

        with click.progressbar(tags_to_update_list, label=f'Writing tags in {len(tags_to_update_list)} files') as bar:
            for t in bar:
                new_tags = {}
                for tag in t.tags_keys:
                    new = t.source_tags[tag] if tag in t.source_tags else ""
                    new_tags[tag] = new
                spoty.audio_files.write_audio_file_tags(t.dest_tags['SPOTY_FILE_NAME'], new_tags)

        context.summary.append(f'  {len(tags_to_update_list)} audio files edited.')

    click.echo('\n------------------------------------------------------------')
    click.echo('\n'.join(context.summary))



from spoty import settings
from spoty import log
from spoty.utils import SpotyContext
import spoty.utils
import click


@click.command("print")
@click.option('--grouping-pattern', '--gp', show_default=True,
              default=settings.SPOTY.DEFAULT_GROUPING_PATTERN,
              help='Tracks will be grouped to playlists according to this pattern.')
@click.option('--print-pattern', '--pp', show_default=True,
              default=settings.SPOTY.DEFAULT_PRINT_PATTERN,
              help='Print a list of tracks according to this formatting pattern.')
@click.pass_obj
def print_tracks(context: SpotyContext,
                 grouping_pattern,
                 print_pattern,
                 ):
    """
Print a list of tracks to console.
    """

    for i, tags_list in enumerate(context.tags_lists):
        if len(context.tags_lists) > 1:
            click.echo()
            click.echo(
                f'============================= LIST {i+1}/{len(context.tags_lists)} =============================')
            click.echo()

        spoty.utils.print_tags_list_grouped(tags_list, print_pattern, grouping_pattern)

        grouped_tags = spoty.utils.group_tags_by_pattern(tags_list, grouping_pattern)

        if len(context.tags_lists) == 1:
            if len(grouped_tags) == 0:
                context.summary.append(f'Total {len(tags_list)} tracks listed.')
            else:
                context.summary.append(
                    f'Total {len(tags_list)} tracks listed (grouped into {len(grouped_tags)} playlists).')
        else:
            if len(grouped_tags) == 0:
                context.summary.append(f'List {i+1}: Total {len(tags_list)} tracks listed.')
            else:
                context.summary.append(
                    f'List {i+1}: Total {len(tags_list)} tracks listed (grouped into {len(grouped_tags)} playlists).')

    click.echo('\n------------------------------------------------------------')
    click.echo('\n'.join(context.summary))



from spoty.commands.first_list_commands import \
    count_command, \
    delete_command, \
    copy_command, \
    move_command, \
    export_command, \
    import_deezer_command, \
    import_spotify_command, \
    sync_spotify_command, \
    find_deezer_group, \
    find_spotify_group, \
    print_command, \
    find_duplicates_command,\
    create_m3u8_command
from spoty.commands import get_second_group
from spoty.utils import SpotyContext
from spoty import settings
import spoty.utils
import click


@click.group("filter")
# @click.option('--playlist-names', '--p', multiple=True,
#               help='Leave only playlists whose names matches this regex filter')
@click.option('--leave-have-tags', '--lht', multiple=True,
              help='Leave only tracks that have all of the specified tags.')
@click.option('--leave-no-tags', '--lnt', multiple=True,
              help='Leave only tracks that do not have any of the specified tags.')
@click.option('--remove-duplicates', '--rd', '-d', is_flag=True,
              help='Remove duplicates.')
@click.option('--leave-duplicates', '--ld', '-l', is_flag=True,
              help='Leave only duplicates.')
@click.option('--duplicates-compare-tags', '--dct', show_default=True, multiple=True,
              default=settings.SPOTY.COMPARE_TAGS_DEFINITELY_DUPLICATE,
              help='Compare duplicates by this tags. It is optional. You can also change the list of tags in the config file.')
@click.option('--leave-added-before', '--lab',
              help='Leave only added to playlist before specified date.')
@click.option('--leave-added-after', '--lab',
              help='Leave only added to playlist after specified date.')
@click.pass_obj
def filter_tracks(context: SpotyContext,
                  # playlist_names,
                  leave_have_tags,
                  leave_no_tags,
                  remove_duplicates,
                  leave_duplicates,
                  duplicates_compare_tags,
                  leave_added_before,
                  leave_added_after,
                  ):
    """
Filter tracks.
    """

    filter_tracks_wrapper(context,
                          # playlist_names,
                          leave_have_tags,
                          leave_no_tags,
                          remove_duplicates,
                          leave_duplicates,
                          duplicates_compare_tags,
                          leave_added_before,
                          leave_added_after,
                          )


def filter_tracks_wrapper(context: SpotyContext,
                          # playlist_names,
                          leave_have_tags,
                          leave_no_tags,
                          remove_duplicates,
                          leave_duplicates,
                          duplicates_compare_tags,
                          leave_added_before,
                          leave_added_after,
                          ):
    tags_list = context.tags_lists[-1]  # get last tags list

    if len(tags_list) > 0:
        summary = []
        summary.append("Filtering:")

        if len(leave_have_tags) > 0:
            new_tags_list = spoty.utils.filter_tags_list_have_tags(tags_list, leave_have_tags)
            if len(tags_list) - len(new_tags_list) != 0:
                summary.append(
                    f'  {len(tags_list) - len(new_tags_list)}/{len(tags_list)} tracks removed (have all of the specified tags).')
            tags_list = new_tags_list

        if len(leave_no_tags) > 0:
            new_tags_list = spoty.utils.filter_tags_list_have_no_tags(tags_list, leave_no_tags)
            if len(tags_list) - len(new_tags_list) != 0:
                summary.append(
                    f'  {len(tags_list) - len(new_tags_list)}/{len(tags_list)} tracks removed (not have any of the specified tags).')
            tags_list = new_tags_list

        if remove_duplicates:
            compare_tags_list = spoty.utils.tuple_to_list(duplicates_compare_tags)
            for compare_tags_str in compare_tags_list:
                compare_tags = compare_tags_str.split(',')
                new_tags_list, dup = spoty.utils.remove_duplicated_tags(tags_list, compare_tags, False)
                if len(dup) > 0:
                    summary.append(
                        f'  {len(dup)}/{len(tags_list)} tracks removed (duplicates by tags: {compare_tags_str})')
                tags_list = new_tags_list

        if leave_duplicates:
            compare_tags_list = spoty.utils.tuple_to_list(duplicates_compare_tags)
            all_dup = []
            for compare_tags_str in compare_tags_list:
                compare_tags = compare_tags_str.split(',')
                new_tags_list, dup = spoty.utils.remove_duplicated_tags(tags_list, compare_tags, False)
                if len(dup) > 0:
                    summary.append(
                        f'   {len(dup)}/{len(tags_list)} tracks left (duplicates by tags: {compare_tags_str})')
                tags_list = new_tags_list
                all_dup.extend(dup)
            tags_list = all_dup

        if leave_added_after:
            new_tags_list = spoty.utils.filter_added_after_date(tags_list, leave_added_after)
            if len(tags_list) - len(new_tags_list) != 0:
                summary.append(
                    f'  {len(tags_list) - len(new_tags_list)}/{len(tags_list)} tracks removed (not added after specified date).')
            tags_list = new_tags_list

        if leave_added_before:
            new_tags_list = spoty.utils.filter_added_before_date(tags_list, leave_added_before)
            if len(tags_list) - len(new_tags_list) != 0:
                summary.append(
                    f'  {len(tags_list) - len(new_tags_list)}/{len(tags_list)} tracks removed (not added before specified date).')
            tags_list = new_tags_list


        if len(summary) > 1:
            context.summary.extend(summary)

    context.tags_lists[-1] = tags_list


filter_tracks.add_command(count_command.count_tracks)
filter_tracks.add_command(print_command.print_tracks)
filter_tracks.add_command(export_command.export_tracks)
filter_tracks.add_command(create_m3u8_command.export_tracks)
filter_tracks.add_command(import_spotify_command.import_spotify)
filter_tracks.add_command(import_deezer_command.import_deezer)
filter_tracks.add_command(sync_spotify_command.sync_spotify)
filter_tracks.add_command(find_deezer_group.find_deezer)
filter_tracks.add_command(find_spotify_group.find_spotify)
filter_tracks.add_command(get_second_group.get_second)
filter_tracks.add_command(delete_command.delete_tracks)
filter_tracks.add_command(move_command.move_tracks)
filter_tracks.add_command(copy_command.copy_tracks)
filter_tracks.add_command(find_duplicates_command.find_duplicates)



from spoty.commands.first_list_commands import \
    count_command, \
    export_command, \
    print_command
from spoty.commands.second_list_commands import \
    filter_second_group, \
    find_duplicates_second_command,\
    find_deezer_second_group, \
    find_spotify_second_group
from spoty.commands import get_group
from spoty.utils import SpotyContext
import click


@click.group("get")
@click.option('--spotify-playlist', '--sp', multiple=True,
              help='Get tracks from Spotify playlist URI or ID.')
@click.option('--spotify-entire-library', '--s', multiple=True,
              help='Get all tracks from Spotify library (by user URI or ID). To request a list for the current authorized user, use "me" as ID.')
@click.option('--spotify-entire-library-regex', '--sr', nargs=2, multiple=True,
              help='Works the same as --spotify-entire-library, but you can specify regex filter which will be applied to playlists names. This way you can query any playlists by names.')
@click.option('--deezer-playlist', '--dp', multiple=True,
              help='Get tracks from Deezer playlist URI or ID.')
@click.option('--deezer-entire-library', '--d', multiple=True,
              help='Get all tracks from Deezer library (by user URI or ID). To request a list for the current authorized user, use "me" as ID.')
@click.option('--deezer-entire-library-regex', '--dr', nargs=2, multiple=True,
              help='Works the same as --deezer-entire-library, but you can specify regex filter which will be applied to playlists names. This way you can query any playlists by names.')
@click.option('--audio', '--a', multiple=True,
              help='Get audio files located at the specified local path. You can specify the audio file name as well.')
@click.option('--csv', '--c', multiple=True,
              help='Get tracks from csv playlists located at the specified local path. You can specify the scv file name as well.')
@click.option('--m3u8', '--m', multiple=True,
              help='Get tracks from m3u8 playlists located at the specified local path. You can specify the m3u8 file name as well.')
@click.option('--no-recursive', '-r', is_flag=True,
              help='Do not search in subdirectories from the specified path.')
@click.pass_obj
def get_second(context: SpotyContext,
               spotify_playlist,
               spotify_entire_library,
               spotify_entire_library_regex,
               deezer_playlist,
               deezer_entire_library,
               deezer_entire_library_regex,
               audio,
               csv,
               m3u8,
               no_recursive
               ):
    """
Collect second list of tracks for further actions (see next commands).
    """

    context.summary.append("Collecting second list:")
    get_group.get_tracks_wrapper(context,
                                 spotify_playlist,
                                 spotify_entire_library,
                                 spotify_entire_library_regex,
                                 deezer_playlist,
                                 deezer_entire_library,
                                 deezer_entire_library_regex,
                                 audio,
                                 csv,
                                 m3u8,
                                 no_recursive,
                                 )


get_second.add_command(filter_second_group.filter_second)

get_second.add_command(count_command.count_tracks)
get_second.add_command(print_command.print_tracks)
get_second.add_command(export_command.export_tracks)
get_second.add_command(find_duplicates_second_command.find_duplicates_second)
get_second.add_command(find_deezer_second_group.find_deezer)
get_second.add_command(find_spotify_second_group.find_spotify)


from spoty import settings
from spoty import log
import spoty.deezer_api
import spoty.csv_playlist
import spoty.utils
from spoty.utils import SpotyContext
import click
import os
from datetime import datetime


@click.command("import-deezer")
@click.option('--grouping-pattern', '--gp', show_default=True,
              default=settings.SPOTY.DEFAULT_GROUPING_PATTERN,
              help='Tracks will be grouped to playlists according to this pattern.')
@click.option('--append', '-a', is_flag=True,
              help='Add tracks to an existing playlist if already exists. If this option is not specified, a new playlist will always be created.')
@click.option('--overwrite', '-o', is_flag=True,
              help='Overwrite existing playlist')
@click.option('--confirm', '-y', is_flag=True,
              help='Confirm all questions with a positive answer automatically.')
@click.option('--ignore-duration', '-i', is_flag=True,
              help='Dont match track duration')
@click.option('--export-result', '-r', is_flag=True,
              help='Export csv files with result (imported, not found, skipped tracks)')
@click.option('--result-path', '--rp',
              default=settings.SPOTY.DEFAULT_EXPORT_PATH,
              help='Path to create resulting csv files')
@click.pass_obj
def import_deezer(context: SpotyContext,
                  grouping_pattern,
                  append,
                  overwrite,
                  confirm,
                  ignore_duration,
                  export_result,
                  result_path
                  ):
    """
Import track list to Deezer library
    """

    for tags_list in context.tags_lists:

        if append and overwrite:
            click.echo(f'Simultaneous use of "--append" and "--overwrite" is not possible',
                       err=True)
            exit()

        click.echo('Next playlists will be imported to Deezer library:')
        grouped_tags = spoty.utils.group_tags_by_pattern(tags_list, grouping_pattern)
        for group_name, g_tags_list in grouped_tags.items():
            click.echo("  " + group_name)
        click.echo(f'Total {len(tags_list)} tracks in {len(grouped_tags)} playlists.')

        if not confirm:
            if click.confirm(f'Do you want to continue?'):
                click.echo("")  # for new line
            else:
                click.echo("\nCanceled.")
                exit()

        found_tags_list, not_found_tags_list = spoty.deezer_api.find_missing_track_ids(tags_list, ignore_duration)

        playlist_ids, imported_tags_list, source_duplicates_tags_list, already_exist_tags_list = \
            spoty.deezer_api.import_playlists_from_tags_list(
                found_tags_list, grouping_pattern, overwrite, append, False, confirm)

        # create result csv playlists

        if export_result:
            result_path = os.path.abspath(result_path)
            date_time_str = datetime.now().strftime("%Y_%m_%d-%H_%M_%S")
            result_path = os.path.join(result_path, 'import-deezer-' + date_time_str)

            if len(imported_tags_list) > 0:
                path = os.path.join(result_path, 'imported')
                spoty.csv_playlist.create_csvs(imported_tags_list, path, grouping_pattern)
            if len(source_duplicates_tags_list) > 0:
                path = os.path.join(result_path, 'skipped_source_duplicates')
                spoty.csv_playlist.create_csvs(source_duplicates_tags_list, path, grouping_pattern)
            if len(already_exist_tags_list) > 0:
                path = os.path.join(result_path, 'skipped_already_exist')
                spoty.csv_playlist.create_csvs(already_exist_tags_list, path, grouping_pattern)
            if len(not_found_tags_list) > 0:
                path = os.path.join(result_path, 'not_found')
                spoty.csv_playlist.create_csvs(not_found_tags_list, path, grouping_pattern)

        # print summery

        context.summary.append("Importing to Deezer:")
        if len(source_duplicates_tags_list) > 0:
            context.summary.append(f'  {len(source_duplicates_tags_list)} duplicates in collected tracks skipped.')
        if len(already_exist_tags_list) > 0:
            context.summary.append(
                f'  {len(already_exist_tags_list)} tracks already exist in Deezer playlists and skipped.')
        if len(not_found_tags_list) > 0:
            context.summary.append(f'  {len(not_found_tags_list)} tracks not found by tags.')

        if len(imported_tags_list) == 0:
            context.summary.append(f'  No tracks to import.')
        else:
            if len(playlist_ids) == 1:
                context.summary.append(f'  {len(imported_tags_list)} tracks imported in Deezer playlist.')
            else:
                context.summary.append(
                    f'  {len(imported_tags_list)} tracks imported in {len(playlist_ids)} Deezer playlists.')

    click.echo('\n------------------------------------------------------------')
    click.echo('\n'.join(context.summary))



from spoty import settings
from spoty import log
from spoty.utils import SpotyContext
import spoty.utils
import spoty.csv_playlist
import click
import os
from datetime import datetime


@click.command("export")
@click.option('--path', '--p', show_default=True,
              default=settings.SPOTY.DEFAULT_EXPORT_PATH,
              help='The path on disk where to export csv files.')
@click.option('--no-source', '-s', is_flag=True,
              help='Do not include source tracks')
@click.option('--no-split-groups', '-g', is_flag=True,
              help='Do not add empty line between groups')
@click.pass_obj
def export_duplicates(context: SpotyContext,
                      path,
                      no_source,
                      no_split_groups
                      ):
    """
Export a list of duplicates to csv file.
    """

    # export result to  csv files
    path = os.path.abspath(path)
    date_time_str = datetime.now().strftime("%Y_%m_%d-%H_%M_%S")
    file_name = 'duplicates-' + date_time_str

    duplicates_count = 0

    all_tags_list = []
    for group in context.duplicates_groups:
        if not no_source:
            all_tags_list.append(group.source_tags)
        all_tags_list.extend(group.def_duplicates)
        all_tags_list.extend(group.prob_duplicates)
        duplicates_count += len(group.def_duplicates)
        duplicates_count += len(group.prob_duplicates)
        if not no_split_groups:
            all_tags_list.append({})

    spoty.csv_playlist.create_csvs(all_tags_list, path, file_name)

    context.summary.append("Exporting:")
    if len(all_tags_list) > 0:
        context.summary.append(
            f'  {duplicates_count} duplicates exported to csv file ("{os.path.join(path, file_name + ".csv")}").')
    else:
        context.summary.append(f'  No tracks to export.')
    click.echo('\n------------------------------------------------------------')
    click.echo('\n'.join(context.summary))



from spoty import settings
from spoty import log
import spoty.csv_playlist
import spoty.utils
from spoty.utils import SpotyContext
import click
import os
from datetime import datetime


@click.command("export")
@click.option('--grouping-pattern', '--gp', show_default=True,
              default=settings.SPOTY.DEFAULT_GROUPING_PATTERN,
              help='Tracks will be grouped to playlists according to this pattern.')
@click.option('--no-duplicates', '-D', is_flag=True,
              help='Prevent duplicates (remove tracks that are already exist in csv file).')
@click.option('--append', '-a', is_flag=True,
              help='Add tracks to an existing csv file if already exists. If this option is not specified, a new csv file will always be created.')
@click.option('--overwrite', '-o', is_flag=True,
              help='Overwrite existing csv file')
@click.option('--path', '--p', show_default=True,
              default=settings.SPOTY.DEFAULT_EXPORT_PATH,
              help='The path on disk where to export csv files.')
@click.option('--no-timestamp', '-T', is_flag=True,
              help='Do not create a subfolder with the current date and time for saved csv files.')
@click.option('--duplicates-compare-tags', '--dct', show_default=True, multiple=True,
              default=settings.SPOTY.COMPARE_TAGS_DEFINITELY_DUPLICATE,
              help='Compare duplicates by this tags. It is optional. You can also change the list of tags in the config file.')
@click.option('--confirm', '-y', is_flag=True,
              help='Confirm all questions with a positive answer automatically.')
@click.pass_obj
def export_tracks(context: SpotyContext,
                  grouping_pattern,
                  no_duplicates,
                  append,
                  overwrite,
                  path,
                  no_timestamp,
                  duplicates_compare_tags,
                  confirm,
                  ):
    """
Export a list of tracks to csv files (playlists) on disk.
    """

    for tags_list in context.tags_lists:

        if append and overwrite:
            click.echo(f'Simultaneous use of "--append" and "--overwrite" is not possible',
                       err=True)
            exit()

        path = os.path.abspath(path)

        if not no_timestamp:
            now = datetime.now()
            date_time_str = now.strftime("%Y_%m_%d-%H_%M_%S")
            path = os.path.join(path, "export-csv-" + date_time_str)

        file_names, names, added_tracks, import_duplicates, already_exist \
            = spoty.csv_playlist.create_csvs(tags_list, path, grouping_pattern, overwrite, append, no_duplicates, confirm,
                                             duplicates_compare_tags)

        # print summery

        context.summary.append("Exporting:")
        if len(import_duplicates) > 0:
            context.summary.append(f'  {len(import_duplicates)} duplicates in collected tracks skipped.')
        if len(already_exist) > 0:
            context.summary.append(f'  {len(already_exist)} tracks already exist in csv files and skipped.')

        if len(added_tracks) == 0:
            context.summary.append(f'  No tracks to export.')
        else:
            if len(file_names) == 1:
                context.summary.append(f'  {len(added_tracks)} tracks written to csv file (path: "{path}").')
            else:
                context.summary.append(
                    f'  {len(added_tracks)} tracks written to {len(file_names)} csv files (path: "{path}").')

    click.echo('\n------------------------------------------------------------')
    click.echo('\n'.join(context.summary))



import spoty.spotify_api
import spoty.audio_files
import click
import spoty.utils


@click.group()
def track():
    r"""Playlists management."""
    pass


@track.command("id")
@click.argument("id")
def track_find_by_id(id):
    r"""
    Find track by ID.
    """
    track = spoty.spotify_api.find_track_by_id(id)
    if track == None:
        click.echo("Not found")
        return
    tags = spoty.spotify_api.read_tags_from_spotify_tracks([track])
    spoty.utils.print_main_tags(tags[0])


@track.command("isrc")
@click.argument("isrc")
def track_find_by_isrc(isrc):
    r"""
    Find track by ISRC.

    Examples:

        spoty spotify track ISRC UK6821402425

    """
    tracks = spoty.spotify_api.find_track_by_isrc(isrc)
    if tracks == None:
        click.echo("Not found")
        return
    for i, track in enumerate(tracks):
        tags = spoty.spotify_api.read_tags_from_spotify_track(track)
        print(f"----------------- {i+1}\{len(tracks)} -------------------")
        spoty.utils.print_main_tags(tags)


@track.command("artist-title")
@click.argument("artist")
@click.argument("title")
def track_find_by_title(artist, title):
    r"""
    Find track by artist and title.

    Examples:

        spoty spotify track artist-title "Aaron Static" "When We Love"

    """
    tracks = spoty.spotify_api.find_track_by_artist_and_title(artist, title)
    if tracks == None:
        click.echo("Not found")
        return
    for i, track in enumerate(tracks):
        tags = spoty.spotify_api.read_tags_from_spotify_track(track)
        print(f"----------------- {i+1}\{len(tracks)} -------------------")
        spoty.utils.print_main_tags(tags)


@track.command("query")
@click.argument("query")
def tracks_find_by_query(query):
    r"""
    Find tracks by query.

    Examples:

        spoty spotify track query "track: breathe"

    """
    tracks = spoty.spotify_api.find_track_by_query(query)
    if tracks == None:
        click.echo("Not found")
        return
    for i, track in enumerate(tracks):
        tags = spoty.spotify_api.read_tags_from_spotify_track(track)
        print(f"----------------- {i+1}\{len(tracks)} -------------------")
        spoty.utils.print_main_tags(tags)



import spoty.utils
from spoty import settings_file_name, secrets_file_name, plugins_path
from spoty.commands import spotify_playlist_commands
from spoty.commands import spotify_like_commands
from spoty.commands import spotify_track_commands
from spoty.commands import get_group
from spoty.commands import deezer_commands
from spoty import spotify_api
from spoty import plugins_path
import click
import os


class SpotyPluginsCLI(click.MultiCommand):

    def list_commands(self, ctx):
        rv = []
        for dir in os.listdir(plugins_path):
            plugin_dir=os.path.abspath(os.path.join(plugins_path,dir))
            if os.path.isdir(plugin_dir):
                filename = os.path.join(plugin_dir,dir+'.py')
                if os.path.isfile(filename):
                    rv.append(dir.replace('_', '-'))
        rv.sort()
        return rv

    def get_command(self, ctx, name):
        ns = {}
        fn = os.path.join(plugins_path, name.replace('-', '_'), name.replace('-', '_') + '.py')

        if not os.path.isfile(fn):
            click.echo(
                f'Plugin "{name}" not installed. Execute "spoty plug" command to see a list of installed plugins.',
                err=True)
            exit()

        with open(fn) as f:
            code = compile(f.read(), fn, 'exec')
            eval(code, ns, ns)
        return ns[name.replace('-', '_')]



@click.group()
def cli():
    """This program allows you to perform various actions with spotify from the console."""
    pass


@cli.command("config")
def config():
    """
Prints configuration parameters.
    """
    click.echo(f'Settings file name: {settings_file_name}')
    click.echo(f'Secrets file name: {secrets_file_name}')
    click.echo(f'Plugins path: {plugins_path}')

@cli.command(cls=SpotyPluginsCLI, help='Plugins.')
def plug():
    pass


cli.add_command(get_group.get_tracks)
cli.add_command(deezer_commands.deezer)


@cli.group()
def spotify():
    """Spotify specific commands."""
    pass


@spotify.command("me")
def spotify_me():
    """Print current user name and id."""
    me = spotify_api.get_sp().me()
    click.echo(f'Current user: "{me["display_name"]}" ({me["id"]})')


spotify.add_command(spotify_playlist_commands.playlist)
spotify.add_command(spotify_like_commands.like)
spotify.add_command(spotify_track_commands.track)

if __name__ == '__main__':
    cli([
        # 'get','export','--help',
    ])



from spoty import settings
from spoty import log
from spoty.utils import SpotyContext
import spoty.utils
import spoty.spotify_api
import spoty.deezer_api
import spoty.audio_files
import click
import os


@click.command("delete")
@click.option('--grouping-pattern', '--gp', show_default=True,
              default=settings.SPOTY.DEFAULT_GROUPING_PATTERN,
              help='Tracks will be grouped to playlists according to this pattern (for listing only).')
@click.option('--print-pattern', '--pp', show_default=True,
              default=settings.SPOTY.DEFAULT_PRINT_PATTERN,
              help='Print a list of tracks according to this formatting pattern.')
# @click.option('--export-result', '-r', is_flag=True,
#               help='Export csv files with result (list of deleted tracks)')
# @click.option('--result-path', '--rp',
#               default=settings.SPOTY.DEFAULT_EXPORT_PATH,
#               help='Path to create resulting csv files')
@click.option('--confirm', '-y', is_flag=True,
              help='Do not ask for confirmation')
@click.pass_obj
def delete_tracks(context: SpotyContext,
                  grouping_pattern,
                  print_pattern,
                  # export_result,
                  # result_path,
                  confirm
                  ):
    """
Delete tracks.
    """
    context.summary.append('Deleting:')

    all_tags_list = []
    for i, tags_list in enumerate(context.tags_lists):
        all_tags_list.extend(tags_list)

    if len(all_tags_list) == 0:
        context.summary.append("  No tracks to delete.")
    else:
        click.echo('Next tracks will be deleted:')
        for i, tags_list in enumerate(context.tags_lists):
            if len(context.tags_lists) > 1:
                click.echo()
                click.echo(
                    f'============================= LIST {i + 1}/{len(context.tags_lists)} =============================')
                click.echo()

            spoty.utils.print_tags_list_grouped(tags_list, print_pattern, grouping_pattern)

        if not confirm:
            click.confirm(f'Are you sure you want to delete {len(all_tags_list)} tracks?', abort=True)



        deleted_spotify_tracks, deleted_deezer_tracks, deleted_audio_files, deleted_csv_tracks = \
            delete_tracks_from_all_sources(all_tags_list)

        if len(deleted_spotify_tracks) > 0:
            context.summary.append(f'  {len(deleted_spotify_tracks)} tracks deleted from Spotify.')

        if len(deleted_deezer_tracks) > 0:
            context.summary.append(f'  {len(deleted_deezer_tracks)} tracks deleted from Deezer.')

        if len(deleted_audio_files) > 0:
            context.summary.append(f'  {len(deleted_audio_files)} audio files deleted.')

        if len(deleted_csv_tracks) > 0:
            context.summary.append(f'  {len(deleted_csv_tracks)} tracks from csv deleted.')

    click.echo('\n------------------------------------------------------------')
    click.echo('\n'.join(context.summary))


def delete_tracks_from_all_sources(all_tags_list):
    spotify_playlists, deezer_playlists, local_audio_files, csv_playlists = spoty.utils.sort_tracks_by_source(all_tags_list)

    # todo: delete from csv

    deleted_spotify_tracks = []
    if len(spotify_playlists.keys()) > 0:
        for playlist_id, track_ids in spotify_playlists.items():
            spoty.spotify_api.remove_tracks_from_playlist(playlist_id, track_ids)
            deleted_spotify_tracks.extend(track_ids)

    deleted_deezer_tracks = []
    if len(deezer_playlists.keys()) > 0:
        for playlist_id, track_ids in deezer_playlists.items():
            spoty.deezer_api.remove_tracks_from_playlist(playlist_id, track_ids)
            deleted_deezer_tracks.extend(track_ids)

    deleted_audio_files = []
    if len(local_audio_files) > 0:
        for file_name in local_audio_files:
            if spoty.utils.is_valid_file(file_name):
                os.remove(file_name)
                deleted_audio_files.append(file_name)

    deleted_csv_tracks = []

    return deleted_spotify_tracks, deleted_deezer_tracks, deleted_audio_files, deleted_csv_tracks



from spoty import settings
from spoty import log
import spoty.spotify_api
import spoty.csv_playlist
import spoty.utils
from spoty.utils import SpotyContext
import click
import os
from datetime import datetime


@click.command("sync-spotify")
@click.option('--grouping-pattern', '--gp', show_default=True,
              default=settings.SPOTY.DEFAULT_GROUPING_PATTERN,
              help='Tracks will be grouped to playlists according to this pattern.')
@click.option('--confirm', '-y', is_flag=True,
              help='Confirm all questions with a positive answer automatically.')
@click.option('--ignore-duration', '-i', is_flag=True,
              help='Dont match track duration')
@click.option('--export-result', '-r', is_flag=True,
              help='Export csv files with result (imported, not found, skipped tracks)')
@click.option('--result-path', '--rp', show_default=True,
              default=settings.SPOTY.DEFAULT_EXPORT_PATH,
              help='Path to create resulting csv files')
@click.option('--source-playlist-prefix', '--spp',
              help='Source playlist names prefix. Only source playlists with names starting with the specified string will be synchronized.')
@click.option('--playlist-prefix', '--pp', show_default=True,
              default=settings.SPOTY.DEFAULT_SYNC_PLAYLIST_PREFIX,
              help='Spotify playlist names prefix. Only Spotify playlists with names starting with the specified string will be synchronized.')
@click.pass_obj
def sync_spotify(context: SpotyContext,
                 grouping_pattern,
                 confirm,
                 ignore_duration,
                 export_result,
                 result_path,
                 source_playlist_prefix,
                 playlist_prefix
                 ):
    """
Sync tracks to Spotify playlists.

\b
When synchronizing, new playlists will be created that are not in the library.
Playlists that already exist will be updated (new tracks are added and non-existent ones are removed).
If a playlist exists in Spotify, but is not present in the source, then such a playlist will be removed from Spotify.
That is why it is advisable to use some kind of prefix in the names of the playlists that will be syncronized.
To do this, use the option --playlist-prefix.
When you use a --playlist-prefix, only playlists with names starting with this prefix will be updated or deleted as a result of synchronization.
If you leave the prefix empty, as a result, all playlists in the library will participate in synchronization, and may be deleted if playlists with the same names are not in the source list.
To avoid accidentally deleting any playlists, the prefix is set by default in the config file. You can override it there, or with --playlist-prefix option.

\b
Playlist names are generated based on the --grouping-pattern.
By default, the pattern is set in the configuration file, but you can override it using the --grouping-pattern option.
For example, if you set the pattern as "%SPOTY_PLAYLIST_NAME%", then the folder name will be used as the playlist name for local audio files.
You can read more about --grouping-pattern in the general Spoty help (spoty --help).

\b
There is also an option --source-playlist-prefix. By default, this option is not used.
By specifying this option, you will filter the playlists in the source track list, leaving only those that start with the specified prefix.
In addition, this prefix will be stripped from the source playlist name during synchronization.
This can be used to sync playlists between different music services or different accounts.

\b
Let's look at an example.
Initially, you had local files and grouped them into playlists by genre using the following command:
  spoty get --a "./music" sync-spotify --gp "%GENRE%" --pp "#SYNC "
As a result, you got the following spotify playlists:
"#SYNC Pop", "#SYNC Rock", "#SYNC Jazz"
Next, you wanted to sync these playlists with another account. Logged into another account, and did sync as follows:
  spoty get --s "MY_FIRST_ACCOUNT_NAME" sync-spotify --gp "%SPOTY_PLAYLIST_NAME%" --pps "#SYNC " --pp "#SYNC "
As a result of this command, you will get an exact copy of the playlists with the same names in another account.
If you remove --pps option, you will get "#SYNC #SYNC Pop", "#SYNC #SYNC Rock", "#SYNC #SYNC Jazz" playlists in the second account.
    """

    if source_playlist_prefix is None:
        source_playlist_prefix = ""
    if playlist_prefix is None:
        playlist_prefix = ""

    for tags_list in context.tags_lists:

        found_tags_list, not_found_tags_list = spoty.spotify_api.find_missing_track_ids(tags_list, ignore_duration)

        click.echo('Next playlists will be synced to Spotify library:')
        grouped_tags = spoty.utils.group_tags_by_pattern(found_tags_list, grouping_pattern)

        new_groups = {}
        for group_name, tags_l in grouped_tags.items():
            if group_name.startswith(source_playlist_prefix):
                new_group_name = group_name[len(source_playlist_prefix):]
                new_group_name = playlist_prefix + new_group_name
                new_groups[new_group_name] = tags_l
        grouped_tags = new_groups

        for group_name, g_tags_list in grouped_tags.items():
            click.echo("  " + group_name)
        click.echo(f'Total {len(found_tags_list)} tracks in {len(grouped_tags)} playlists.')

        if not confirm:
            if click.confirm(f'\nDo you want to continue?'):
                click.echo("")  # for new line
            else:
                click.echo("\nCanceled.")
                exit()

        playlist_ids, imported_tags_list, source_duplicates_tags_list, already_exist_tags_list = \
            spoty.spotify_api.import_playlists_from_tag_groups(grouped_tags, True, False, False, confirm)

        deleted_playlists=[]
        playlists = spoty.spotify_api.get_list_of_playlists()
        for playlist in playlists:
            if playlist['name'].startswith(playlist_prefix):
                if playlist['name'] not in grouped_tags.keys():
                    res = spoty.spotify_api.delete_playlist(playlist['id'], confirm)
                    if res:
                        deleted_playlists.append(playlist)

        # create result csv playlists

        if export_result:
            result_path = os.path.abspath(result_path)
            date_time_str = datetime.now().strftime("%Y_%m_%d-%H_%M_%S")
            result_path = os.path.join(result_path, 'import-spotify-' + date_time_str)

            if len(imported_tags_list) > 0:
                path = os.path.join(result_path, 'added')
                spoty.csv_playlist.create_csvs(imported_tags_list, path, playlist_prefix + grouping_pattern)
            if len(source_duplicates_tags_list) > 0:
                path = os.path.join(result_path, 'skipped_source_duplicates')
                spoty.csv_playlist.create_csvs(source_duplicates_tags_list, path, playlist_prefix + grouping_pattern)
            if len(already_exist_tags_list) > 0:
                path = os.path.join(result_path, 'skipped_already_exist')
                spoty.csv_playlist.create_csvs(already_exist_tags_list, path, playlist_prefix + grouping_pattern)
            if len(not_found_tags_list) > 0:
                path = os.path.join(result_path, 'not_found')
                spoty.csv_playlist.create_csvs(not_found_tags_list, path, playlist_prefix + grouping_pattern)

        # print summery

        context.summary.append("Sync to Spotify:")
        if len(source_duplicates_tags_list) > 0:
            context.summary.append(f'  {len(source_duplicates_tags_list)} duplicates in collected tracks skipped.')
        if len(already_exist_tags_list) > 0:
            context.summary.append(f'  {len(already_exist_tags_list)} tracks already exist in playlists and skipped.')
        if len(not_found_tags_list) > 0:
            context.summary.append(f'  {len(not_found_tags_list)} tracks not found by tags.')
        if len(deleted_playlists) > 0:
            context.summary.append(f'  {len(deleted_playlists)} playlists deleted.')

        if len(imported_tags_list) == 0:
            context.summary.append(f'  No new tracks found.')
        else:
            context.summary.append(
                    f'  {len(imported_tags_list)} new tracks added.')

    click.echo('\n------------------------------------------------------------')
    click.echo('\n'.join(context.summary))



from spoty.commands.first_list_commands import \
    count_command, \
    export_command, \
    print_command
from spoty.commands.second_list_commands import \
    find_duplicates_second_command
from spoty import settings
import spoty.spotify_api
import spoty.csv_playlist
import spoty.utils
from spoty.utils import SpotyContext
import click
import os
from datetime import datetime


@click.group("find-spotify")
@click.option('--grouping-pattern', '--gp', show_default=True,
              default=settings.SPOTY.DEFAULT_GROUPING_PATTERN,
              help='Tracks will be grouped to playlists according to this pattern (only for exporting csv result files).')
@click.option('--ignore-duration', '-i', is_flag=True,
              help='Dont match track duration')
@click.option('--export-result', '-r', is_flag=True,
              help='Export csv files with result (imported, not found, skipped tracks)')
@click.option('--result-path', '--rp',
              default=settings.SPOTY.DEFAULT_EXPORT_PATH,
              help='Path to create resulting csv files')
@click.pass_obj
def find_spotify(context: SpotyContext,
                 grouping_pattern,
                 ignore_duration,
                 export_result,
                 result_path
                 ):
    """
Find tracks from second list in Spotify.
    """

    tags_list = context.tags_lists[1]
    tags_with_ids = []
    tags_without_ids = []
    for tags in tags_list:
        if 'SPOTIFY_TRACK_ID' in tags:
            tags_with_ids.append(tags)
        else:
            tags_without_ids.append(tags)

    found_tags_list, not_found_tags_list = spoty.spotify_api.find_missing_track_ids(tags_list, ignore_duration)

    context.tags_lists[1] = found_tags_list

    # create result csv playlists

    if export_result:
        result_path = os.path.abspath(result_path)
        date_time_str = datetime.now().strftime("%Y_%m_%d-%H_%M_%S")
        result_path = os.path.join(result_path, 'import-spotify-' + date_time_str)

        if len(found_tags_list) > 0:
            path = os.path.join(result_path, 'found')
            spoty.csv_playlist.create_csvs(found_tags_list, path, grouping_pattern)
        if len(not_found_tags_list) > 0:
            path = os.path.join(result_path, 'not_found')
            spoty.csv_playlist.create_csvs(not_found_tags_list, path, grouping_pattern)

    # print summery

    context.summary.append("Finding in Spotify:")
    if len(tags_with_ids) > 0:
        context.summary.append(f'  {len(tags_with_ids)} tracks have SPOTIFY_TRACK_ID tag and skipped.')
    if len(found_tags_list) > 0:
        context.summary.append(f'  {len(found_tags_list)} tracks found.')
    if len(not_found_tags_list) > 0:
        context.summary.append(f'  {len(not_found_tags_list)} tracks not found.')


find_spotify.add_command(count_command.count_tracks)
find_spotify.add_command(print_command.print_tracks)
find_spotify.add_command(export_command.export_tracks)
find_spotify.add_command(find_duplicates_second_command.find_duplicates_second)


from spoty.commands.first_list_commands import \
    count_command, \
    export_command, \
    print_command, \
    filter_group
from spoty.commands.second_list_commands import \
    find_duplicates_second_command,\
    find_deezer_second_group,\
    find_spotify_second_group
from spoty.utils import SpotyContext
from spoty import settings
import click


@click.group("filter")
# @click.option('--playlist-names', '--p', multiple=True,
#               help='Leave only playlists whose names matches this regex filter')
@click.option('--leave-have-tags', '--lht', multiple=True,
              help='Leave only tracks that have all of the specified tags.')
@click.option('--leave-no-tags', '--lnt', multiple=True,
              help='Leave only tracks that do not have any of the specified tags.')
@click.option('--remove-duplicates', '--rd', '-d', is_flag=True,
              help='Remove duplicates.')
@click.option('--leave-duplicates', '--ld', '-l', is_flag=True,
              help='Leave only duplicates.')
@click.option('--duplicates-compare-tags', '--dct', show_default=True, multiple=True,
              default=settings.SPOTY.COMPARE_TAGS_DEFINITELY_DUPLICATE,
              help='Compare duplicates by this tags. It is optional. You can also change the list of tags in the config file.')
@click.option('--leave-added-before', '--lab',
              help='Leave only added to playlist before specified date.')
@click.option('--leave-added-after', '--lab',
              help='Leave only added to playlist after specified date.')
@click.pass_obj
def filter_second(context: SpotyContext,
                  # playlist_names,
                  leave_have_tags,
                  leave_no_tags,
                  remove_duplicates,
                  leave_duplicates,
                  duplicates_compare_tags,
    leave_added_before,
    leave_added_after,
                  ):
    """
Filter tracks.
    """

    filter_group.filter_tracks_wrapper(context,
                                       # playlist_names,
                                       leave_have_tags,
                                       leave_no_tags,
                                       remove_duplicates,
                                       leave_duplicates,
                                       duplicates_compare_tags,
                                       leave_added_before,
                                       leave_added_after,
                                       )


filter_second.add_command(count_command.count_tracks)
filter_second.add_command(print_command.print_tracks)
filter_second.add_command(export_command.export_tracks)
filter_second.add_command(find_duplicates_second_command.find_duplicates_second)
filter_second.add_command(find_deezer_second_group.find_deezer)
filter_second.add_command(find_spotify_second_group.find_spotify)



from spoty.commands.first_list_commands import count_command
from spoty.commands.duplicates_commands import \
    add_missing_tags_command, \
    replace_tags_command, \
    export_duplicates_command, \
    print_duplicates_command, \
    get_duplicates_group, \
    get_unique_second_group
from spoty.utils import SpotyContext
from spoty import settings
from spoty import log
import spoty.deezer_api
import spoty.csv_playlist
import spoty.deezer_api
import spoty.spotify_api
import spoty.audio_files
import spoty.utils
import click
import os
from datetime import datetime


@click.group("dup")
@click.option('--compare-tags-def', '--ctd', show_default=True, multiple=True,
              default=settings.SPOTY.COMPARE_TAGS_DEFINITELY_DUPLICATE,
              help='Compare definitely duplicates by this tags. It is optional. You can also change the list of tags in the config file.')
@click.option('--compare-tags-prob', '--ctp', show_default=True, multiple=True,
              default=settings.SPOTY.COMPARE_TAGS_PROBABLY_DUPLICATE,
              help='Compare probably duplicates by this tags. It is optional. You can also change the list of tags in the config file.')
@click.option('--only-def', '--def', '-d', is_flag=True,
              help='Find only definitely duplicates.')
@click.option('--only-prob', '--prob', '-p', is_flag=True,
              help='Find only probably duplicates.')
@click.pass_obj
def find_duplicates_second(context: SpotyContext,
                           compare_tags_def,
                           compare_tags_prob,
                           only_def,
                           only_prob
                           ):
    """
Find duplicates between the first and second list of tracks.
    """

    if only_def and only_prob:
        click.echo(f'Simultaneous use of "--only-def" and "--only-prob" is not possible', err=True)
        exit()

    source_list = context.tags_lists[0]
    dest_list = context.tags_lists[1]

    compare_tags_def = spoty.utils.tuple_to_list(compare_tags_def)
    compare_tags_prob = spoty.utils.tuple_to_list(compare_tags_prob)

    if only_def:
        compare_tags_prob = []
    if only_prob:
        compare_tags_def = []

    duplicates_groups, unique_source_tracks, unique_dest_tracks, sources_def_dups, sources_prob_dups = \
        spoty.utils.find_duplicates_in_tag_lists(source_list, dest_list, compare_tags_def, compare_tags_prob,
                                                 True, True)

    context.duplicates_groups = duplicates_groups
    context.unique_first_tracks = unique_source_tracks
    context.unique_second_tracks = unique_dest_tracks

    total_def_duplicates_count = 0
    total_prob_duplicates_count = 0

    for group in context.duplicates_groups:
        total_def_duplicates_count += len(group.def_duplicates)
        total_prob_duplicates_count += len(group.prob_duplicates)

    context.summary.append("Finding duplicates:")
    if len(unique_source_tracks) > 0:
        context.summary.append(f'  {len(unique_source_tracks)} unique tracks in first list found.')
    if len(sources_def_dups) > 0:
        context.summary.append(f'  {len(sources_def_dups)} definitely duplicates in first list found.')
    if len(sources_prob_dups) > 0:
        context.summary.append(f'  {len(sources_prob_dups)} probably duplicates in first list found.')
    if len(unique_dest_tracks) > 0:
        context.summary.append(f'  {len(unique_dest_tracks)} unique tracks in second list found.')
    if total_def_duplicates_count > 0:
        context.summary.append(f'  {total_def_duplicates_count} definitely duplicates in second list found.')
    if total_prob_duplicates_count > 0:
        context.summary.append(f'  {total_prob_duplicates_count} probably duplicates in second list found.')

    if total_def_duplicates_count == 0 and total_prob_duplicates_count == 0:
        context.summary.append(f'  0 duplicates found.')


find_duplicates_second.add_command(count_command.count_tracks)
find_duplicates_second.add_command(add_missing_tags_command.add_missing_tags)
find_duplicates_second.add_command(replace_tags_command.replace_tags)
find_duplicates_second.add_command(export_duplicates_command.export_duplicates)
find_duplicates_second.add_command(print_duplicates_command.print_duplicates)
find_duplicates_second.add_command(get_duplicates_group.get_duplicates)
find_duplicates_second.add_command(get_unique_second_group.get_unique)



from spoty.commands.first_list_commands import \
    count_command, \
    delete_command, \
    copy_command, \
    move_command, \
    export_command, \
    import_deezer_command, \
    import_spotify_command, \
    sync_spotify_command, \
    print_command, \
    create_m3u8_command
from spoty.utils import SpotyContext
import click


@click.group("get")
@click.option('--only-def', '--od', '-d', is_flag=True,
              help='Get only definitely duplicates.')
@click.option('--only-prob', '--op', '-p', is_flag=True,
              help='Get only probably duplicates.')
@click.option('--add-source', '--as', '-s', is_flag=True,
              help='Get source tracks too.')
@click.pass_obj
def get_duplicates(context: SpotyContext,
                   only_def,
                   only_prob,
add_source
                   ):
    """
Get duplicates for further actions (see next commands).
    """

    if only_def and only_prob:
        click.echo(f'Simultaneous use of "--only-def" and "--only-prob" is not possible', err=True)
        exit()

    context.tags_lists.clear()
    context.tags_lists.append([])

    for group in context.duplicates_groups:
        if add_source:
            context.tags_lists[0].append(group.source_tags)
        if not only_prob:
            context.tags_lists[0].extend(group.def_duplicates)
        if not only_def:
            context.tags_lists[0].extend(group.prob_duplicates)

    context.summary.append("Collecting duplicates:")
    context.summary.append(f'  {len(context.tags_lists[0])} duplicates collected.')

    context.duplicates_groups = []
    context.unique_first_tracks = []
    context.unique_second_tracks = []

get_duplicates.add_command(count_command.count_tracks)
get_duplicates.add_command(print_command.print_tracks)
get_duplicates.add_command(export_command.export_tracks)
get_duplicates.add_command(create_m3u8_command.export_tracks)
get_duplicates.add_command(import_spotify_command.import_spotify)
get_duplicates.add_command(import_deezer_command.import_deezer)
get_duplicates.add_command(sync_spotify_command.sync_spotify)
get_duplicates.add_command(delete_command.delete_tracks)
get_duplicates.add_command(move_command.move_tracks)
get_duplicates.add_command(copy_command.copy_tracks)




from spoty import settings
from spoty import log
import spoty.deezer_api
import spoty.csv_playlist
import spoty.deezer_api
import spoty.spotify_api
import spoty.audio_files
import spoty.utils
from spoty.commands import get_group
from spoty.utils import SpotyContext
from spoty.utils import DuplicatesGroup
import click
import os
from datetime import datetime


@click.command("add-missing-tags")
@click.option('--tags', '--t',
              help='Which tags to search (specify separated by commas). If not specified, will search all tags.')
@click.option('--ignore-tags', '--T',
              help='Which tags to ignore (specify separated by commas).',
              default=settings.SPOTY.DEFAULT_IGNORE_MISSING_TAGS)
@click.option('--confirm', '-y', is_flag=True,
              help='Do not ask for confirmation.')
@click.option('--from-second-too', '-s', is_flag=True,
              help='Collect tags from tracks in second list too (from duplicates).')
@click.option('--to-first-too', '-f', is_flag=True,
              help='Write missing tags to tracks from first list too.')
@click.pass_obj
def add_missing_tags(context: SpotyContext,
                     tags,
                     ignore_tags,
                     confirm,
                     from_second_too,
                     to_first_too
                     ):
    """
Collect tags from all duplicated tracks and add missing tags (update them all).
    """
    context.summary.append('Adding missing tags:')

    compare_tags = tags.split(',') if tags is not None else []
    for i in range(len(compare_tags)):
        compare_tags[i] = compare_tags[i].upper()

    ignore_tags = ignore_tags.split(',') if ignore_tags is not None else []
    for i in range(len(ignore_tags)):
        ignore_tags[i] = ignore_tags[i].upper()

    tags_to_add = {}
    with click.progressbar(context.duplicates_groups, label='Collecting missing tags') as bar:
        for group in bar:
            # collect all tags
            all_group_tags = group.source_tags.copy()
            if from_second_too:
                for tags in group.def_duplicates:
                    new_tags = spoty.utils.get_missing_tags(all_group_tags, tags, compare_tags,ignore_tags)
                    for key, value in new_tags.items():
                        all_group_tags[key] = value
                for tags in group.prob_duplicates:
                    new_tags = spoty.utils.get_missing_tags(all_group_tags, tags, compare_tags,ignore_tags)
                    for key, value in new_tags.items():
                        all_group_tags[key] = value

            # find missing tags
            if to_first_too:
                if 'SPOTY_FILE_NAME' in group.source_tags:  # is local file
                    new_tags = spoty.utils.get_missing_tags(group.source_tags, all_group_tags, compare_tags,ignore_tags)
                    if len(new_tags.keys()) > 0:
                        tags_to_add[group.source_tags['SPOTY_FILE_NAME']] = new_tags
            for tags in group.def_duplicates:
                if 'SPOTY_FILE_NAME' in tags:  # is local file
                    new_tags = spoty.utils.get_missing_tags(tags, all_group_tags, compare_tags,ignore_tags)
                    if len(new_tags.keys()) > 0:
                        tags_to_add[tags['SPOTY_FILE_NAME']] = new_tags
            for tags in group.prob_duplicates:
                if 'SPOTY_FILE_NAME' in tags:  # is local file
                    new_tags = spoty.utils.get_missing_tags(tags, all_group_tags, compare_tags,ignore_tags)
                    if len(new_tags.keys()) > 0:
                        tags_to_add[tags['SPOTY_FILE_NAME']] = new_tags

    if len(tags_to_add.items()) == 0:
        context.summary.append("  No missing tags found")
    else:

        click.echo('Next audio files will be edited:')

        for file_name, tags in tags_to_add.items():
            click.echo(f'  {file_name}:\n    {",".join(tags.keys())}')

        if not confirm:
            click.confirm(f'Are you sure you want to edit tags in {len(tags_to_add.items())} audio files?', abort=True)

        with click.progressbar(tags_to_add.items(), label=f'Writing tags in {len(tags_to_add.items())} files') as bar:
            for file_name, tags in bar:
                spoty.audio_files.write_audio_file_tags(file_name, tags)

        context.summary.append(f'  {len(tags_to_add.items())} audio files edited.')

    click.echo('\n------------------------------------------------------------')
    click.echo('\n'.join(context.summary))



from spoty import log
import spoty.utils
import spoty.audio_files
import click
import os
import csv
import time
import re


class CSVImportException(Exception):
    """Base class for exceptions when importing CSV file."""
    pass


class CSVFileEmpty(CSVImportException):
    """File is empty."""
    pass


class CSVFileInvalidHeader(CSVImportException):
    """The header of csv table does not contain any of the required fields."""
    pass


def is_csv(file_name):
    return file_name.upper().endswith('.CSV')


def create_csvs(tags_list, path, grouping_pattern, overwrite=False, append=False, remove_duplicates=False,
                confirm=False, compare_tags_list=None):
    path = os.path.abspath(path)

    if compare_tags_list is None:
        compare_tags_list = []

    all_created_csv_file_names = []
    all_created_csv_names = []
    all_added_tags = []
    all_import_duplicates = []
    all_already_exist = []

    if len(tags_list) > 0:
        grouped_tags = spoty.utils.group_tags_by_pattern(tags_list, grouping_pattern)

        for group, tags_l in grouped_tags.items():
            csv_name = spoty.utils.slugify_file_pah(group)
            csv_file_name = os.path.join(path, csv_name + '.csv')

            create_new_file = True
            if os.path.isfile(csv_file_name):
                if append:
                    create_new_file = False
                elif overwrite:
                    if not confirm:
                        if not click.confirm(f'File "{csv_file_name}" already exist. Overwrite?'):
                            click.echo("\nCanceled")
                            continue
                        click.echo("")  # for new line
                else:
                    csv_file_name = spoty.utils.find_empty_file_name(csv_file_name)

            if remove_duplicates:
                for compare_tags_str in compare_tags_list:
                    compare_tags = compare_tags_str.split(',')
                    tags_l, import_duplicates = spoty.utils.remove_duplicated_tags(tags_l, compare_tags, False)
                    all_import_duplicates.extend(import_duplicates)

                if len(all_import_duplicates) > 0:
                    log.debug(f'{len(all_import_duplicates)} duplicates found when adding tracks. It will be skipped.')

                if not create_new_file:
                    exist_tags_list = read_tags_from_csv(csv_file_name)
                    for compare_tags_str in compare_tags_list:
                        compare_tags = compare_tags_str.split(',')
                        tags_l, exist_tags = spoty.utils.remove_exist_tags(exist_tags_list, tags_l, compare_tags, False)
                        all_already_exist.extend(exist_tags)

            write_tags_to_csv(tags_l, csv_file_name, not create_new_file)

            all_created_csv_names.append(csv_name)
            all_created_csv_file_names.append(csv_file_name)
            all_added_tags.extend(tags_l)

    return all_created_csv_file_names, all_created_csv_names, all_added_tags, all_import_duplicates, all_already_exist


def find_csvs_in_paths(paths, recursive=True):
    all_csvs = []

    for path in paths:
        file_names = find_csvs_in_path(path, recursive)
        all_csvs.extend(file_names)

    return all_csvs


def find_csvs_in_path(path, recursive=True):
    path = os.path.abspath(path)

    if not spoty.utils.is_valid_path(path):
        raise Exception("Path is not valid: " + path)

    full_file_names = []
    if recursive:
        full_file_names = [os.path.join(dp, f) for dp, dn, filenames in os.walk(path) for f in filenames if
                           is_csv(os.path.splitext(f)[1])]
    else:
        full_file_names = [os.path.join(path, f) for f in os.listdir(path) if os.path.isfile(os.path.join(path, f))]
        full_file_names = list(filter(lambda f: is_csv(f), full_file_names))

    return full_file_names


def write_tags_to_csv(tags_list, csv_file_name, append=False):
    tags_list = spoty.utils.clean_tags_list_before_write(tags_list)

    for i, tags in enumerate(tags_list):
        for key, value in tags.items():
            if type(value) is str:
                if len(value) > 131072:
                    mess = f'Tag "{key}" has value larger than csv field limit (131072) and will be truncated (file: "{csv_file_name}", line: {i + 1}).'
                    click.echo('\n' + mess)
                    log.warning(mess)
                    tags[key] = value[0: 131071]

    if tags_list is None or len(tags_list) == 0:
        return

    if append:
        if os.path.isfile(csv_file_name):
            old_tags = read_tags_from_csv(csv_file_name, False, False)
            old_tags.extend(tags_list)
            tags_list = old_tags

    # collect all keys
    keys = []
    for tags in tags_list:
        for key, value in tags.items():
            if not key in keys:
                keys.append(key)

    keys = spoty.utils.reorder_tag_keys_main_first(keys)

    # add missing keys
    for tags in tags_list:
        for key in keys:
            if not key in tags:
                tags[key] = ""

    rows = []
    rows.append(keys)  # header
    for tags in tags_list:
        row = [tags[key] for key in keys]
        rows.append(row)

    os.makedirs(os.path.dirname(csv_file_name), exist_ok=True)
    with open(csv_file_name, 'w', encoding='utf-8-sig', newline='') as file:
        writer = csv.writer(file)
        writer.writerows(rows)


def read_tags_from_csvs(csv_file_names, add_spoty_tags=True):
    all_tags_lists = []
    for csv_file_name in csv_file_names:
        tags_list = read_tags_from_csv(csv_file_name, add_spoty_tags)
        all_tags_lists.extend(tags_list)

    return all_tags_lists


def read_tags_from_csv(csv_file_name, add_spoty_tags=True, add_missing_tags=True):
    csv_file_name = os.path.abspath(csv_file_name)
    tags_list = []

    with open(csv_file_name, newline='', encoding='utf-8-sig') as file:
        reader = csv.reader(file)
        if sum(1 for row in reader) == 0:
            raise CSVFileEmpty()

    with open(csv_file_name, newline='', encoding='utf-8-sig') as file:
        header = []
        reader = csv.reader(file)

        for i, row in enumerate(reader):
            # read header

            if (i == 0):
                header = row
                if len(header) == 0:
                    raise CSVFileInvalidHeader()
                continue

            # read tags

            if len(row) == 0:  # skip empty lines
                continue

            if all(item == "" for item in row):  # skip empty lines
                continue

            tags = {}

            for h, key in enumerate(header):
                if len(row[h]) > 0:
                    tags[key] = row[h]

            if add_spoty_tags:
                playlist_name = os.path.basename(csv_file_name)
                if playlist_name.upper().endswith('.CSV'):
                    playlist_name = playlist_name[:-4]
                tags['SPOTY_SOURCE'] = 'CSV'
                tags['SPOTY_PLAYLIST_NAME'] = playlist_name
                tags['SPOTY_PLAYLIST_INDEX'] = str(i)

            if add_missing_tags:
                tags = spoty.utils.clean_tags_after_read(tags)

            tags_list.append(tags)

    return tags_list



import os.path
from datetime import datetime
import click
from spoty import settings
from typing import List
import dateutil.parser
import numpy as np
from multiprocessing import Process, Lock, Queue, Value, Array
import sys
import time
from time import strftime
from time import gmtime

THREADS_COUNT = 12

tag_allies = [
    ['YEAR', 'DATE'],
    ['TRACK', 'TRACKNUMBER'],
    ['DISK', 'DISKNUMBER']
]

spoty_tags = \
    [
        'SPOTY_DUP_GROUP',
        'SPOTY_DEF_DUP_TAGS',
        'SPOTY_PROB_DUP_TAGS',
        'SPOTY_DUP_LIST',
        'SPOTY_DUP_ID',
        'SPOTY_FOUND_BY',
        'SPOTY_SOURCE',
        'SPOTY_PLAYLIST_NAME',
        'SPOTY_PLAYLIST_ID',
        'SPOTY_PLAYLIST_INDEX',
        'SPOTY_FILE_NAME',
        'SPOTY_TRACK_ID',
        'SPOTY_TRACK_ADDED',
        'SPOTY_LENGTH',
        'SPOTY_TRACK_LISTENED',
    ]

spotify_tags = [
    'SPOTIFY_TRACK_ID',
    'SPOTIFY_ALBUM_ID',
]

deezer_tags = [
    'DEEZER_TRACK_ID',
    'DEEZER_ALBUM_ID',
    'DEEZER_ARTIST_ID',
    'DEEZER_LYRICS_ID',
]

main_tags = \
    [
        'ISRC',
        'ARTIST',
        'ALBUMARTIST',
        'TITLE',
        'ALBUM',
        'GENRE',
        'MOOD',
        'OCCASION',
        'RATING',
        'COMMENT'
        'SOURCE'
        'BPM',
        'QUALITY',
        'TEMPO',
        'YEAR',
    ]

additional_tags = \
    [
        '1T_TAGGEDDATE',  # auto tagger
        'AUTHOR',
        'COMPILATION',
        'COMPOSER',
        'COPYRIGHT',
        'DISC',
        'ENCODER',
        'EXPLICIT',
        'FILEOWNER',
        'GAIN',
        'INITIAL KEY',
        'INITIALKEY',
        'ENGINEER',
        'INVOLVEDPEOPLE',
        'ITUNESADVISORY',
        'LABEL',
        'LOVE RATING',
        'LYRICS',
        'MIXER',
        'PRODUCER',
        'PUBLISHER',
        'REPLAYGAIN_TRACK_GAIN',
        'RELEASE DATE',
        'STYLE',
        'TOTALDISCS',
        'TOTALTRACKS',
        'TRACK',
        'UPC',
        'WRITER',
    ]


class DuplicatesGroup:
    source_tags: dict
    def_duplicates: list
    prob_duplicates: list
    def_found_tags: list
    prob_found_tags: list

    def __init__(self):
        self.source_tags = {}
        self.def_duplicates = []
        self.prob_duplicates = []
        self.def_found_tags = []
        self.prob_found_tags = []

    def get_duplicates_count(self):
        return len(self.def_duplicates) + len(self.prob_duplicates)

    def has_duplicates(self):
        return self.get_duplicates_count() > 0


class SpotyContext:
    tags_lists: list
    summary: list
    duplicates_groups: List[DuplicatesGroup]
    unique_first_tracks: list
    unique_second_tracks: list

    def __init__(self):
        self.tags_lists = []
        self.summary = []
        self.duplicates_groups = []
        self.unique_first_tracks = []
        self.unique_second_tracks = []


mutex = Lock()


def tuple_to_list(some_tuple: tuple):
    l = []
    l.extend(some_tuple)
    return l


def dict_to_list(some_dics: dict):
    l = []
    for key, value in some_dics.items():
        l.append(value)
    return l


def is_valid_path(path: str):
    return os.path.isdir(path)


def is_valid_file(path: str):
    return os.path.isfile(path)


def slugify_file_pah(text: str):
    # valid_chars = "-_.()=!@#$%%^&+ %s%s" % (string.ascii_letters, string.digits)
    # return ''.join(c for c in text if c in valid_chars).strip()

    valid_chars = '<>:"/\|?*'

    for char in valid_chars:
        text = text.replace(char, '')

    return text


def filter_duplicates(src_arr: list, dest_arr: list):
    return list(filter(lambda id: id not in src_arr, dest_arr))


def remove_duplicates(arr: list):
    good = []
    duplicates = []
    for item in arr:
        if item in good:
            duplicates.append(item)
        else:
            good.append(item)
    return good, duplicates


def remove_exist(exist_arr: list, new_arr: list):
    new = []
    exist = []
    for item in new_arr:
        if item in exist_arr:
            exist.append(item)
        else:
            new.append(item)
    return new, exist


def remove_duplicated_tags(tags_list: list, tags_to_compare: list, allow_missing=False, show_progressbar=False):
    good = []
    duplicates = []
    if show_progressbar:
        bar = click.progressbar(length=len(tags_list), label=f'Finding duplicates in {len(tags_list)} tracks')

    for new_tags in tags_list:
        if show_progressbar:
            bar.update(1)

        found = False
        for exist_tags in good:
            if compare_tags(exist_tags, new_tags, tags_to_compare, allow_missing):
                duplicates.append(new_tags)
                found = True
                break
        if not found:
            good.append(new_tags)

    if show_progressbar:
        bar.finish()
        click.echo()

    return good, duplicates


def remove_exist_tags(exist_tags_list: list, new_tags_list: list, tags_to_compare: list, allow_missing=False,
                      show_progressbar=False):
    new = []
    exist = []
    if show_progressbar:
        bar = click.progressbar(new_tags_list,
                                label=f'Searching for tags matching in {len(exist_tags_list)} and {len(new_tags_list)} tracks')

    for new_tags in new_tags_list:
        if show_progressbar:
            bar.update(1)

        found = False
        for exist_tags in exist_tags_list:
            if compare_tags(exist_tags, new_tags, tags_to_compare, allow_missing):
                exist.append(new_tags)
                found = True
                break
        if not found:
            new.append(new_tags)

    if show_progressbar:
        bar.finish()
        click.echo()

    return new, exist


def remove_exist_tags_by_isrc_and_length(exist_tags_list: list, new_tags_list: list, show_progressbar=False):
    new = []
    exist = []
    if show_progressbar:
        bar = click.progressbar(new_tags_list,
                                label=f'Searching for tags matching in {len(exist_tags_list)} and {len(new_tags_list)} tracks')

    COMPARE_LENGTH_TOLERANCE_SEC = int(settings.SPOTY.COMPARE_LENGTH_TOLERANCE_SEC)

    exist_tags_dict = {}
    for tags in exist_tags_list:
        if 'ISRC' in tags and 'SPOTY_LENGTH' in tags:
            if tags['ISRC'] not in exist_tags_dict:
                exist_tags_dict[tags['ISRC']] = []
            exist_tags_dict[tags['ISRC']].append(tags['SPOTY_LENGTH'])

    for new_tags in new_tags_list:
        if show_progressbar:
            bar.update(1)

        found = False

        if 'ISRC' in new_tags and 'SPOTY_LENGTH' in new_tags:
            if new_tags['ISRC'] in exist_tags_dict:
                for exist_length in exist_tags_dict[new_tags['ISRC']]:
                    if abs(int(new_tags['SPOTY_LENGTH']) - int(exist_length) < COMPARE_LENGTH_TOLERANCE_SEC):
                        exist.append(new_tags)
                        found = True
        if not found:
            new.append(new_tags)

    if show_progressbar:
        bar.finish()
        click.echo()

    return new, exist


def compare_tags(tags1: dict, tags2: dict, tags_to_compare: list, allow_missing=False):
    for tag in tags_to_compare:

        if not tag in tags1 or not tag in tags2:
            if allow_missing:
                continue
            else:
                return False

        if tag == 'SPOTY_LENGTH':
            if abs(int(tags1['SPOTY_LENGTH']) - int(tags2['SPOTY_LENGTH'])) \
                    > settings.SPOTY.COMPARE_LENGTH_TOLERANCE_SEC:
                return False
            else:
                continue

        if tag == "ARTIST":
            artist1 = tags1[tag].replace(',', ';').upper()
            artist1 = artist1.split(';')
            artist2 = tags2[tag].replace(',', ';').upper()
            artist2 = artist2.split(';')
            found = False
            for art in artist1:
                if art in artist2:
                    found = True
            if not found:
                return False
            else:
                continue

        if tag == "TITLE":
            title1 = tags1[tag].upper()
            title1 = ''.join(char for char in title1 if char.isalnum())
            title2 = tags2[tag].upper()
            title2 = ''.join(char for char in title2 if char.isalnum())
            if not title2.startswith(title1) and not title1.startswith(title2):
                return False
            else:
                continue

        if tag == "ALBUM":
            album1 = tags1[tag].upper()
            album2 = tags2[tag].upper()
            if not album2.startswith(album1) and not album1.startswith(album2):
                return False
            else:
                continue

        if tag == "ISRC":
            isrc1 = tags1[tag].upper().replace('-', '')
            isrc2 = tags2[tag].upper().replace('-', '')
            if isrc1 != isrc2:
                return False
            else:
                continue

        if tags1[tag] != tags2[tag]:
            return False

    return True


def find_duplicates_in_tags(tags_list: list, compare_tags: list):
    if len(compare_tags) == 0:
        return

    duplicates = {}
    pattern = ""
    for tag in compare_tags:
        pattern += "%" + tag + "%,"
    pattern = pattern[:-1]

    groupped_tags = group_tags_by_pattern(tags_list, pattern, "Unknown")

    for group, tags in groupped_tags.items():
        if group == "Unknown":
            continue
        if len(tags) > 1:
            if not group in duplicates:
                duplicates[group] = []
            duplicates[group].extend(tags)

    skipped_tags = groupped_tags['Unknown'] if 'Unknown' in groupped_tags else []

    return duplicates, skipped_tags


def print_main_tags(tags: dict):
    if 'ISRC' in tags: print(f'ISRC: {tags["ISRC"]}')
    if 'ARTIST' in tags: print(f'ARTIST: {tags["ARTIST"]}')
    if 'TITLE' in tags: print(f'TITLE: {tags["TITLE"]}')
    if 'ALBUM' in tags: print(f'ALBUM: {tags["ALBUM"]}')
    if 'GENRE' in tags: print(f'GENRE: {tags["GENRE"]}')
    if 'MOOD' in tags: print(f'MOOD: {tags["MOOD"]}')
    if 'OCCASION' in tags: print(f'OCCASION: {tags["OCCASION"]}')
    if 'RATING' in tags: print(f'RATING: {tags["RATING"]}')
    if 'COMMENT' in tags: print(f'COMMENT: {tags["COMMENT"]}')
    if 'BARCODE' in tags: print(f'BARCODE: {tags["BARCODE"]}')
    if 'SPOTY_LENGTH' in tags:
        seconds = int(tags["SPOTY_LENGTH"])
        m, s = divmod(seconds, 60)
        time = '{:02d}:{:02d}'.format(m, s)
        print(f'SPOTY_LENGTH: {tags["SPOTY_LENGTH"]} ({time})')
    if 'SPOTIFY_TRACK_ID' in tags: print(f'SPOTIFY_TRACK_ID: {tags["SPOTIFY_TRACK_ID"]}')
    if 'DEEZER_TRACK_ID' in tags: print(f'DEEZER_TRACK_ID: {tags["DEEZER_TRACK_ID"]}')
    if 'SOURCE' in tags: print(f'SOURCE: {tags["SOURCE"]}')
    if 'SOURCEID' in tags: print(f'SOURCEID: {tags["SOURCEID"]}')
    if 'YEAR' in tags: print(f'YEAR: {tags["YEAR"]}')


def print_tags_list_grouped(tags_list: list, print_pattern: str, grouping_pattern: str):
    if len(tags_list) == 0:
        return

    grouped_tags = group_tags_by_pattern(tags_list, grouping_pattern)

    for group, tags_l in grouped_tags.items():
        print(f'\n------------------------- {group}:')
        print_tags_list(tags_l, print_pattern)


def print_tags_list(tags_list: list, print_pattern: str):
    if len(tags_list) == 0:
        return

    for tags in tags_list:
        txt = parse_pattern(tags, print_pattern)
        print("  " + txt)


def print_duplicates_tags_list(tags_list: list, print_pattern: str = None):
    if len(tags_list) == 0:
        return

    for tags in tags_list:
        if print_pattern is None:
            print_pattern = settings.DUPLICATE_PRINT_PATTERN[tags['SPOTY_SOURCE']]
        txt = parse_pattern(tags, print_pattern)
        print("  " + txt)


def check_tag_has_allies(tag: str):
    for allies in tag_allies:
        if tag in allies:
            return True
    return False


def get_tag_allies(tag: str, include_source_tag=True):
    res = []
    for allies in tag_allies:
        if tag in allies:
            res = allies.copy()

    if tag in res:
        res.remove(tag)

    if include_source_tag:
        res.append(tag)
    return res


def print_tags(tags: dict, tags_to_print: list):
    for tag in tags_to_print:
        allies = get_tag_allies(tag, True)
        for a in allies:
            if a.upper() in tags:
                print(f'{a}: {tags[a]}')


def add_playlist_index_from_playlist_names(tags_list: list):
    res = []
    groups = group_tags_by_pattern(tags_list, "%SPOTY_PLAYLIST_NAME%")
    for group, g_tags_list in groups.items():
        for i, tags in enumerate(g_tags_list):
            tags['SPOTY_PLAYLIST_INDEX'] = str(i + 1)
            res.append(tags)
    return res


def filter_tags_list_have_tags(tags_list: list, filter_tags: list):
    filtered = []
    for tags in tags_list:
        if check_all_tags_exist(tags, filter_tags):
            filtered.append(tags)
    return filtered


def filter_tags_list_have_no_tags(tags_list: list, filter_tags: list):
    filtered = []
    for tags in tags_list:
        if not check_all_tags_exist(tags, filter_tags):
            filtered.append(tags)
    return filtered


def filter_added_after_date(tags_list: list, date: str, add_if_date_tag_missing=False):
    filtered = []
    for tags in tags_list:
        if 'SPOTY_TRACK_ADDED' in tags:
            track_added = datetime.strptime(tags['SPOTY_TRACK_ADDED'], "%Y-%m-%d %H:%M:%S")
            # specified_date = datetime.strptime(added_after_time, "%Y-%m-%d %H:%M:%S")
            try:
                specified_date = dateutil.parser.parse(date)
            except:
                click.echo(f'Cant parse date: "{date}". Use this format: "2018-06-29 08:15:27"', err=True)
                exit()
            if track_added > specified_date:
                filtered.append(tags)
        else:
            if add_if_date_tag_missing:
                filtered.append(tags)
    return filtered


def filter_added_before_date(tags_list: list, date: str, add_if_date_tag_missing=False):
    filtered = []
    for tags in tags_list:
        if 'SPOTY_TRACK_ADDED' in tags:
            track_added = datetime.strptime(tags['SPOTY_TRACK_ADDED'], "%Y-%m-%d %H:%M:%S")
            # specified_date = datetime.strptime(added_after_time, "%Y-%m-%d %H:%M:%S")
            try:
                specified_date = dateutil.parser.parse(date)
            except:
                click.echo(f'Cant parse date: "{date}". Use this format: "2018-06-29 08:15:27"', err=True)
                exit()
            if track_added < specified_date:
                filtered.append(tags)
        else:
            if add_if_date_tag_missing:
                filtered.append(tags)
    return filtered


def check_all_tags_exist(tags: dict, tags_to_check: list):
    for tag in tags_to_check:
        if not tag.upper() in tags:
            return False
    return True


def group_tags_by_pattern(tags_list: list, pattern: str, not_found_tag_name="Unknown"):
    groups = {}

    for tags in tags_list:
        group_name = parse_pattern(tags, pattern)

        if not group_name in groups:
            groups[group_name] = []

        groups[group_name].append(tags)

    return groups


def parse_pattern(tags: dict, pattern: str):
    result = ""
    tag_name = ""
    building_tag = False
    for c in pattern:
        if c == "%":
            building_tag = not building_tag
            if not building_tag:
                allies = get_tag_allies(tag_name, True)
                for a in allies:
                    if a in tags:
                        tag = tags[a]
                        result += str(tag)
                tag_name = ""
        else:
            if building_tag:
                tag_name += c
                tag_name = tag_name.upper()
            else:
                result += c

    return result


def reorder_tag_keys_main_first(keys: list):
    res = []

    # reorder spoty tags first
    for key in spoty_tags:
        if key in keys:
            res.append(key)

    for key in spotify_tags:
        if key in keys:
            res.append(key)

    for key in deezer_tags:
        if key in keys:
            res.append(key)

    # reorder main tags first
    for key in main_tags:
        if key in keys:
            res.append(key)

    # add other tags
    for key in keys:
        if not key in res:
            res.append(key)

    return res


def get_missing_tags(exist_tags: dict, new_tags: dict, compare_tags: list = None, ignore_tags: list = None):
    if compare_tags is None:
        compare_tags = []

    if ignore_tags is None:
        ignore_tags = []

    missing_tags = {}

    for key, value in new_tags.items():
        if len(compare_tags) > 0:
            if key not in compare_tags:
                continue

        if len(ignore_tags) > 0:
            if key in ignore_tags:
                continue

        if key == 'LENGTH':
            continue

        if key in spoty_tags:
            continue

        if key in exist_tags:
            continue

        found = False
        for aliases in tag_allies:
            if key in aliases:
                for al in aliases:
                    if al in exist_tags:
                        found = True
        if found:
            continue

        missing_tags[key] = value

    return missing_tags


def find_empty_file_name(exist_file_name: str):
    exist_file_name = os.path.abspath(exist_file_name)

    if not os.path.isfile(exist_file_name):
        return exist_file_name

    base_name = os.path.basename(exist_file_name)
    ext = os.path.splitext(base_name)[1]
    base_name = os.path.splitext(base_name)[0]
    dir_name = os.path.dirname(exist_file_name)
    i = 1
    while True:
        i += 1
        new_file_name = os.path.join(dir_name, base_name + f' {i}' + ext)
        if not os.path.isfile(new_file_name):
            return new_file_name


def clean_tags_list_before_write(tags_list):
    for tags in tags_list:
        if 'SPOTY_PLAYLIST_INDEX' in tags:
            del tags['SPOTY_PLAYLIST_INDEX']
        if 'LENGTH' in tags:
            del tags['LENGTH']
    return tags_list


def clean_tags_list_after_read(tags_list):
    for i, tags in enumerate(tags_list):
        tags_list[i] = clean_tags_after_read(tags)


def clean_tags_after_read(tags):
    # local files from deemix

    if 'ISRC' in tags:
        tags['ISRC'] = tags['ISRC'].upper().replace('-', '')

    if 'SOURCEID' in tags and 'DEEZER_TRACK_ID' not in tags \
            and 'SOURCE' in tags and tags['SOURCE'].upper() == "DEEZER":
        tags['DEEZER_TRACK_ID'] = tags['SOURCEID']

    # missing deezer track id

    if 'SPOTY_SOURCE' in tags and tags['SPOTY_SOURCE'].upper() == "DEEZER":

        if 'SPOTY_TRACK_ID' not in tags and 'DEEZER_TRACK_ID' in tags:
            tags['SPOTY_TRACK_ID'] = tags['DEEZER_TRACK_ID']

        if 'DEEZER_TRACK_ID' not in tags and 'SPOTY_TRACK_ID' in tags:
            tags['DEEZER_TRACK_ID'] = tags['SPOTY_TRACK_ID']

    # missing spotify track id

    if 'SPOTY_SOURCE' in tags and tags['SPOTY_SOURCE'].upper() == "SPOTIFY":

        if 'SPOTY_TRACK_ID' not in tags and 'SPOTIFY_TRACK_ID' in tags:
            tags['SPOTY_TRACK_ID'] = tags['SPOTIFY_TRACK_ID']

        if 'SPOTIFY_TRACK_ID' not in tags and 'SPOTY_TRACK_ID' in tags:
            tags['SPOTIFY_TRACK_ID'] = tags['SPOTY_TRACK_ID']

    return tags


def find_duplicates_in_groups(check_tags: dict, groups: List[DuplicatesGroup], compare_tags_list: list,
                              compare_with_def_duplicates=False, compare_with_prob_duplicates=False) -> (
        DuplicatesGroup, list):
    if len(compare_tags_list) == 0:
        return None, None

    for tags_to_compare in compare_tags_list:
        for group in groups:
            if len(group.source_tags.items()) > 0:
                if compare_tags(check_tags, group.source_tags, tags_to_compare, False):
                    return group, tags_to_compare

    if compare_with_def_duplicates:
        for tags_to_compare in compare_tags_list:
            for group in groups:
                for tags in group.def_duplicates:
                    if compare_tags(check_tags, tags, tags_to_compare, False):
                        return group, tags_to_compare

    if compare_with_prob_duplicates:
        for tags_to_compare in compare_tags_list:
            for group in groups:
                for tags in group.prob_duplicates:
                    if compare_tags(check_tags, tags, tags_to_compare, False):
                        return group, tags_to_compare
    return None, None


def find_duplicates_in_tag_list2(tags_list: list, compare_tags_def_list: list, compare_tags_prob_list: list,
                                 add_dup_tags=False):
    # get tags to compare from config

    for i, tags in enumerate(compare_tags_def_list):
        compare_tags_def_list[i] = tags.split(',')

    for i, tags in enumerate(compare_tags_prob_list):
        compare_tags_prob_list[i] = tags.split(',')

    groups: List[DuplicatesGroup] = []

    # find duplicates

    with click.progressbar(tags_list, label=f'Finding duplicates in {len(tags_list)} tracks') as bar:
        for tags in bar:
            group, found_tags = find_duplicates_in_groups(tags, groups, compare_tags_def_list, True, True)
            if group is not None:
                group.def_duplicates.append(tags)
                group.def_found_tags.append(found_tags)
            else:
                group, found_tags = find_duplicates_in_groups(tags, groups, compare_tags_prob_list, True, True)
                if group is not None:
                    group.prob_duplicates.append(tags)
                    group.prob_found_tags.append(found_tags)
                else:
                    d = DuplicatesGroup()
                    d.source_tags = tags
                    groups.append(d)

    # remove unique

    unique_tracks = []
    duplicates_groups: List[DuplicatesGroup] = []
    for group in groups:
        if group.has_duplicates():
            duplicates_groups.append(group)
        else:
            unique_tracks.append(group.source_tags)

    if add_dup_tags:
        for i, group in enumerate(duplicates_groups):
            if len(group.source_tags.items()) > 0:
                group.source_tags['SPOTY_DUP_GROUP'] = i + 1
            for y, tags in enumerate(group.def_duplicates):
                tags['SPOTY_DUP_GROUP'] = i + 1
                tags['SPOTY_DEF_DUP_TAGS'] = ','.join(group.def_found_tags[y])
            for y, tags in enumerate(group.prob_duplicates):
                tags['SPOTY_DUP_GROUP'] = i + 1
                tags['SPOTY_PROB_DUP_TAGS'] = ','.join(group.prob_found_tags[y])

    return duplicates_groups, unique_tracks


def find_duplicates_in_tag_lists(source_list: list, dest_list: list, compare_tags_def_list: list,
                                 compare_tags_prob_list: list,
                                 add_dup_tags=False, remove_duplicates_in_source=True):
    # get tags to compare from config

    for i, tags in enumerate(compare_tags_def_list):
        compare_tags_def_list[i] = tags.split(',')

    for i, tags in enumerate(compare_tags_prob_list):
        compare_tags_prob_list[i] = tags.split(',')

    # find duplicates in dest

    groups: List[DuplicatesGroup] = []
    unique_dest_tracks = []

    for source_tags in source_list:
        d = DuplicatesGroup()
        d.source_tags = source_tags
        groups.append(d)

    if len(source_list) + len(dest_list) < 2000:  # single thread
        with click.progressbar(dest_list,
                               label=f'Finding duplicates in {len(source_list) + len(dest_list)} tracks') as bar:
            for dest_tags in bar:
                group, found_tags = find_duplicates_in_groups(dest_tags, groups, compare_tags_def_list)
                if group is not None:
                    group.def_duplicates.append(dest_tags)
                    group.def_found_tags.append(found_tags)
                else:
                    group, found_tags = find_duplicates_in_groups(dest_tags, groups, compare_tags_prob_list)
                    if group is not None:
                        group.prob_duplicates.append(dest_tags)
                        group.prob_found_tags.append(found_tags)
                    else:
                        unique_dest_tracks.append(dest_tags)
    else:  # multi thread
        try:
            parts = np.array_split(dest_list, THREADS_COUNT)
            threads = []
            counters = []
            results = Queue()

            with click.progressbar(length=len(dest_list),
                                   label=f'Finding duplicates in {len(source_list) + len(dest_list)} tracks') as bar:

                # start threads
                for i, part in enumerate(parts):
                    counter = Value('i', 0)
                    counters.append(counter)
                    dest_list_part = list(part)
                    thread = Process(target=find_duplicates_in_groups_thread, args=(
                        dest_list_part, groups, compare_tags_def_list, compare_tags_prob_list, counter, results))
                    threads.append(thread)
                    thread.daemon = True  # This thread dies when main thread exits
                    thread.start()

                    # update bar
                    total = sum([x.value for x in counters])
                    added = total - bar.pos
                    if added > 0:
                        bar.update(added)

                # waiting for complete
                while not bar.finished:
                    time.sleep(0.1)
                    total = sum([x.value for x in counters])
                    added = total - bar.pos
                    if added > 0:
                        bar.update(added)

                # combine results
                for i in range(len(parts)):
                    res = results.get()
                    unique_dest_tracks.extend(res['unique_dest_tracks'])
                    for i, group in enumerate(res['groups']):
                        if len(group.def_duplicates) > 0:
                            groups[i].def_duplicates.extend(group.def_duplicates)
                            groups[i].def_found_tags.extend(group.def_found_tags)
                        if len(group.prob_duplicates) > 0:
                            groups[i].prob_duplicates.extend(group.prob_duplicates)
                            groups[i].prob_found_tags.extend(group.prob_found_tags)

        except (KeyboardInterrupt, SystemExit):  # aborted by user
            click.echo()
            click.echo('Aborted.')
            sys.exit()

    # remove unique source
    unique_source_tracks = []
    temp_groups: List[DuplicatesGroup] = []
    for group in groups:
        if group.has_duplicates():
            temp_groups.append(group)
        else:
            unique_source_tracks.append(group.source_tags)
    groups = temp_groups

    # remove duplicates in unique source tracks
    sources_def_dups = []
    sources_prob_dups = []
    if remove_duplicates_in_source:
        unique_sources = []

        with click.progressbar(unique_source_tracks,
                               label=f'Finding duplicates in {len(unique_source_tracks)} source tracks') as bar:
            for dest_tags in bar:
                group, found_tags = find_duplicates_in_groups(dest_tags, groups, compare_tags_def_list)
                if group is not None:
                    sources_def_dups.append(dest_tags)
                else:
                    group, found_tags = find_duplicates_in_groups(dest_tags, groups, compare_tags_prob_list)
                    if group is not None:
                        sources_prob_dups.append(dest_tags)
                    else:
                        unique_sources.append(dest_tags)

        unique_source_tracks = unique_sources

    if add_dup_tags:
        for i, group in enumerate(groups):
            group.source_tags['SPOTY_DUP_GROUP'] = i + 1
            for y, tags in enumerate(group.def_duplicates):
                tags['SPOTY_DUP_GROUP'] = i + 1
                tags['SPOTY_DEF_DUP_TAGS'] = ','.join(group.def_found_tags[y])
            for y, tags in enumerate(group.prob_duplicates):
                tags['SPOTY_DUP_GROUP'] = i + 1
                tags['SPOTY_PROB_DUP_TAGS'] = ','.join(group.prob_found_tags[y])

    return groups, unique_source_tracks, unique_dest_tracks, sources_def_dups, sources_prob_dups


def find_duplicates_in_groups_thread(dest_list, groups, compare_tags_def_list, compare_tags_prob_list, counter, result):
    unique_dest_tracks = []

    for i, dest_tags in enumerate(dest_list):
        group, found_tags = find_duplicates_in_groups(dest_tags, groups, compare_tags_def_list)
        if group is not None:
            group.def_duplicates.append(dest_tags)
            group.def_found_tags.append(found_tags)
        else:
            group, found_tags = find_duplicates_in_groups(dest_tags, groups, compare_tags_prob_list)
            if group is not None:
                group.prob_duplicates.append(dest_tags)
                group.prob_found_tags.append(found_tags)
            else:
                unique_dest_tracks.append(dest_tags)
        if (i + 1) % 10 == 0:
            counter.value += 10
        if i + 1 == len(dest_list):
            counter.value += (i % 10) + 1

    res = {}
    res['unique_dest_tracks'] = unique_dest_tracks
    res['groups'] = groups

    result.put(res)


def compare_by_tags(source_list: list, dest_list: list, tags_to_compare: list, dest_unique: dict, dest_dups: dict,
                    dup_tag: str, add_dup_tags=False):
    unique = []
    dups = []
    for dest_tags in dest_list:
        found = False
        for source_tags in source_list:
            if compare_tags(source_tags, dest_tags, tags_to_compare, False):
                found = True
                if add_dup_tags:
                    if dup_tag not in dest_tags:
                        dest_tags[dup_tag] = ""
                    dest_tags[dup_tag] += f'{source_tags["SPOTY_DUP_ID"]} : {",".join(tags_to_compare)}\n'
        if found:
            dups.append(dest_tags)
        else:
            unique.append(dest_tags)

    # move duplicates from unique to dups
    for item in dups:
        id = item['SPOTY_DUP_ID']
        if id in dest_unique:
            dest_dups[id] = item
            del dest_unique[id]


def move_audio_files_to_path(tags_list, path):
    moved_files = []
    for tags in tags_list:
        if 'SPOTY_FILE_NAME' in tags:
            old_file_name = tags['SPOTY_FILE_NAME']

            base_name = os.path.basename(old_file_name)

            new_file_name = os.path.join(path, base_name)
            if os.path.isfile(new_file_name):
                new_file_name = find_empty_file_name(new_file_name)

            os.rename(old_file_name, new_file_name)
            moved_files.append(new_file_name)

    return moved_files


def sort_tracks_by_source(tags_list):
    spotify_playlists = {}
    deezer_playlists = {}
    local_audio_files = []
    csv_playlists = {}
    for tags in tags_list:
        if tags['SPOTY_SOURCE'] == 'SPOTIFY':
            playlist_id = tags['SPOTY_PLAYLIST_ID']
            if playlist_id not in spotify_playlists:
                spotify_playlists[playlist_id] = []
            spotify_playlists[playlist_id].append(tags['SPOTIFY_TRACK_ID'])

        if tags['SPOTY_SOURCE'] == 'DEEZER':
            playlist_id = tags['SPOTY_PLAYLIST_ID']
            if playlist_id not in deezer_playlists:
                deezer_playlists[playlist_id] = []
            deezer_playlists[playlist_id].append(tags['DEEZER_TRACK_ID'])

        if tags['SPOTY_SOURCE'] == 'LOCAL':
            local_audio_files.append(tags['SPOTY_FILE_NAME'])

        if tags['SPOTY_SOURCE'] == 'CSV':
            playlist_name = tags['SPOTY_PLAYLIST_NAME']
            if playlist_name not in csv_playlists:
                csv_playlists[playlist_name] = []
            csv_playlists[playlist_name].append(tags)

    return spotify_playlists, deezer_playlists, local_audio_files, csv_playlists



from spoty import settings
import spoty
from deezersdk.deezersdk import Deezer
from spoty import settings
import os

DEEZER_APP_ID = settings.default.DEEZER_APP_ID
DEEZER_APP_SECRET = settings.default.DEEZER_APP_SECRET
DEEZER_ACCESS_TOKEN = settings.default.DEEZER_ACCESS_TOKEN
DEEZER_REDIRECT_URI = settings.DEEZER.REDIRECT_URI
arl_file_name = os.path.join(spoty.config_path, '.arl')


def get_token():
    url = Deezer.get_oauth_login_url(
        app_id=DEEZER_APP_ID,
        redirect_uri=DEEZER_REDIRECT_URI
    )

    print("Open this url in browser and copy code from url (...callback?code=).")
    print(url)

    print("Enter code here:")

    code = input()

    token = Deezer.get_oauth_token(app_id=DEEZER_APP_ID, app_secret=DEEZER_APP_SECRET, code=code)

    print("Access token:")
    print(token)






from spoty import settings
import spoty
from flask import Flask, redirect, request
import os
import requests
import webbrowser
import toml


DEEZER_APP_ID = settings.default.DEEZER_APP_ID
DEEZER_APP_SECRET = settings.default.DEEZER_APP_SECRET
DEEZER_ACCESS_TOKEN = settings.default.DEEZER_ACCESS_TOKEN
DEEZER_REDIRECT_URI = settings.DEEZER.REDIRECT_URI
arl_file_name = os.path.join(spoty.config_path, '.arl')

def on_started():
    print('Server is running')
    webbrowser.open("http://localhost:8888")


class AuthFlaskApp(Flask):
    def run(self, host=None, port=None, debug=None, load_dotenv=True, **options):
        if not self.debug or os.getenv('WERKZEUG_RUN_MAIN') == 'true':
            with self.app_context():
                on_started()
        super(AuthFlaskApp, self).run(host=host, port=port, debug=debug, load_dotenv=load_dotenv, **options)


app = AuthFlaskApp(__name__)


def write_token_to_config_file(token):
    data = toml.load(spoty.secrets_file_name)
    data['default']['DEEZER_ACCESS_TOKEN'] = token

    f = open(spoty.secrets_file_name, 'w')
    toml.dump(data, f)
    f.close()

    print("Config file updated.")


def shutdown_server():
    func = request.environ.get('werkzeug.server.shutdown')
    if func is None:
        raise RuntimeError('Not running with the Werkzeug Server')
    func()


@app.route('/', methods=['GET'])
def default():
    url = (f'https://connect.deezer.com/oauth/auth.php?app_id={DEEZER_APP_ID}'
           f'&redirect_uri={DEEZER_REDIRECT_URI}&perms=manage_library,delete_library,email,offline_access')
    print(url)
    return redirect(url, code=302)


@app.route('/callback', methods=['GET'])
def deezer_login():
    code = request.args.get('code')

    url = (f'https://connect.deezer.com/oauth/access_token.php?app_id={DEEZER_APP_ID}'
           f'&secret={DEEZER_APP_SECRET}&code={code}&output=json')

    response = requests.get(url)

    if response.text == 'wrong code':
        return 'wrong code'

    response = response.json()
    token = response['access_token']
    print("Access token: " + str(token))
    write_token_to_config_file(token)


    shutdown_server()

    return f'<h1>Access token received</h1>{token}<br><br>Now you can use spoty.', {'Content-Type': 'text/html'}


def get_token():
    app.run(host='0.0.0.0', port=8888)


if __name__ == '__main__':
    get_token()



from spoty import settings
from spoty import log
import spoty.spotify_api
import spoty.audio_files
import spoty.csv_playlist
import click
import re
import spoty.utils
import time
import os
from datetime import datetime


@click.group()
def playlist():
    r"""Playlists management."""
    pass


@playlist.command("list")
@click.option('--filter-names', '--fn',
              help='List only playlists whose names matches this regex filter')
@click.option('--user-id', '--u',
              help='Get playlists of this user.')
@click.option('--followed', '-f', is_flag=True,
              help='List followed by user playlists instead of created by user (only for current user).')
def playlist_list(filter_names, user_id, followed):
    r"""
    List of all playlists.

    Examples:

        spoty spotify playlist list

        spoty spotify playlist list --user-id 4717400682
    """
    if user_id == None:
        playlists = spoty.spotify_api.get_list_of_playlists(not followed)
    else:
        playlists = spoty.spotify_api.get_list_of_user_playlists(user_id)

    if len(playlists) == 0:
        exit()

    if filter_names is not None:
        playlists = list(filter(lambda pl: re.findall(filter_names, pl['name']), playlists))
        click.echo(f'{len(playlists)} playlists matches the filter')

    if len(playlists) == 0:
        exit()

    for playlist in playlists:
        click.echo(f'{playlist["id"]} "{playlist["name"]}"')

    click.echo(f'Total playlists: {len(playlists)}')


@playlist.command("read")
@click.argument("playlist_ids", nargs=-1)
def playlist_read(playlist_ids):
    r"""
    Read playlists and print track ids.
    it could be your playlists or created by another user.

    PLAYLIST_IDS - list of playlist IDs or URIs. You can specify one ID or many IDs separated by a space.

    Examples:

        spoty spotify playlist read 37i9dQZF1DX8z1UW9HQvSq

        spoty spotify playlist read 37i9dQZF1DX8z1UW9HQvSq 37i9dQZF1DX7jNFrjYQurt

        spoty spotify playlist read https://open.spotify.com/playlist/37i9dQZF1DX8z1UW9HQvSq

    """
    tracks = []
    for playlist_id in playlist_ids:
        click.echo(f'Tracks in playlist {playlist_id}:')
        tracks = spoty.spotify_api.get_tracks_of_playlist(playlist_id)
        for track in tracks:
            title = spoty.spotify_api.get_track_artist_and_title(track["track"])
            click.echo(f'{track["track"]["id"]}: {title}')

    click.echo(f'Total tracks: {len(tracks)}')


@playlist.command("create")
@click.argument("name")
def playlist_create(name):
    r"""
    Create a new empty playlist with the specified name.

    Examples:

        spoty spotify playlist create "My awesome playlist"
    """
    id = spoty.spotify_api.create_playlist(name)
    click.echo(f'New playlist created (id: {id}, name: "{name}")')


@playlist.command("delete")
@click.argument("playlist_ids", nargs=-1)
@click.option('--confirm', '-y', is_flag=True,
              help='Do not ask for confirmation')
def playlist_delete(playlist_ids, confirm):
    r"""
    Delete playlists.

        Examples:

        spoty spotify playlist delete 37i9dQZF1DX8z1UW9HQvSq

        spoty spotify playlist delete 37i9dQZF1DX8z1UW9HQvSq 37i9dQZF1DX7jNFrjYQurt

        spoty spotify playlist delete https://open.spotify.com/playlist/37i9dQZF1DX8z1UW9HQvSq
    """

    if len(playlist_ids) == 0:
        return

    click.echo("Found playlists: ")
    for playlist_id in playlist_ids:
        playlist = spoty.spotify_api.get_playlist(playlist_id)
        if playlist is not None:
            click.echo("  " + playlist['name'])
        else:
            click.echo(f'  Playlist {playlist_id} not found')

    if not confirm:
        click.confirm(f'Are you sure you want to delete {len(playlist_ids)} playlists?', abort=True)

    for playlist_id in playlist_ids:
        res = spoty.spotify_api.delete_playlist(playlist_id, True)
        if res:
            click.echo(f'Playlist {playlist_id} deleted')
        else:
            click.echo(f'Playlist {playlist_id} not deleted')


@playlist.command("delete-all")
@click.option('--filter-names', '--fn',
              help='List only playlists whose names matches this regex filter')
@click.option('--confirm', '-y', is_flag=True,
              help='Do not ask for export confirmation')
def playlist_delete_all(filter_names, confirm):
    r"""
    Delete all playlists in the library.

    Examples:

        spoty spotify playlist delete-all

        spoty spotify playlist delete-all --confirm
    """

    playlists = spoty.spotify_api.get_list_of_playlists()

    if len(playlists) == 0:
        exit()

    if filter_names is not None:
        playlists = list(filter(lambda pl: re.findall(filter_names, pl['name']), playlists))
        click.echo(f'{len(playlists)} playlists matches the filter')

    if len(playlists) == 0:
        exit()

    click.echo("Found playlists: ")
    for playlist in playlists:
        click.echo(f'  {playlist["id"]} "{playlist["name"]}"')

    if not confirm:
        click.confirm(f'Are you sure you want to delete {len(playlists)} playlists?', abort=True)

    with click.progressbar(playlists, label='Deleting playlists') as bar:
        for playlist in bar:
            spoty.spotify_api.delete_playlist(playlist["id"], True)

    click.echo(f'{len(playlists)} playlist deleted')


@playlist.command("copy")
@click.argument("playlist_ids", nargs=-1)
def playlist_copy(playlist_ids):
    r"""
    Create copies of playlists.
    it could be your playlists or created by another user.
    The exact same playlists will be created in your library.

    PLAYLIST_IDS - list of playlist IDs or URIs. You can specify one ID or many IDs separated by a space.



    Examples:

        spoty spotify playlist copy 37i9dQZF1DX8z1UW9HQvSq

        spoty spotify playlist copy 37i9dQZF1DX8z1UW9HQvSq 37i9dQZF1DX7jNFrjYQurt

        spoty spotify playlist copy https://open.spotify.com/playlist/37i9dQZF1DX8z1UW9HQvSq

    """
    playlists = []
    tracks = []
    with click.progressbar(playlist_ids, label='Copying playlists') as bar:
        for playlist_id in bar:
            new_playlist_id, tracks_added = spoty.spotify_api.copy_playlist(playlist_id)
            playlists.extend(new_playlist_id)
            tracks.extend(tracks_added)

    click.echo(f'{len(playlists)} playlists with {len(tracks)} tracks copied.')


@playlist.command("add-tracks")
@click.argument("playlist_id")
@click.argument("track_ids", nargs=-1)
@click.option('--allow-duplicates', '-d', is_flag=True,
              help='Add tracks that are already in the playlist.')
def playlist_add_tracks(playlist_id, track_ids, allow_duplicates):
    r"""
    Add tracks to playlist.

    PLAYLIST_ID - playlist ID or URI.

    TRACK_IDS - list of track IDs or URIs. You can specify one ID or many IDs separated by a space.

    Examples:

        spoty spotify playlist add-tracks 37i9dQZF1DX8z1UW9HQvSq 00i9VF7sjSaTqblAuKFBDO

        spoty spotify playlist add-tracks 37i9dQZF1DX8z1UW9HQvSq 00i9VF7sjSaTqblAuKFBDO 7cjlfruK9Oqw7k5wAZGO72

        spoty spotify playlist add-tracks https://open.spotify.com/playlist/37i9dQZF1DX8z1UW9HQvSq https://open.spotify.com/track/00i9VF7sjSaTqblAuKFBDO

    """
    track_ids = list(track_ids)
    tracks_added, import_duplicates, already_exist, invalid_ids \
        = spoty.spotify_api.add_tracks_to_playlist_by_ids(playlist_id, track_ids, allow_duplicates)

    if len(invalid_ids) > 0:
        click.echo("Invalid track IDs:")
        for id in invalid_ids:
            click.echo(id)
        click.echo('---------------------------')

    click.echo(f'{len(tracks_added)} tracks added to playlist {playlist_id}')


@playlist.command("remove-tracks")
@click.argument("playlist_id")
@click.argument("track_ids", nargs=-1)
def playlist_remove_tracks(playlist_id, track_ids):
    r"""
    Remove tracks from playlist.

    PLAYLIST_ID - playlist ID or URI.

    TRACK_IDS - list of track IDs or URIs. You can specify one ID or many IDs separated by a space.

    Examples:

        spoty spotify playlist remove-tracks 37i9dQZF1DX8z1UW9HQvSq 00i9VF7sjSaTqblAuKFBDO

        spoty spotify playlist remove-tracks 37i9dQZF1DX8z1UW9HQvSq 00i9VF7sjSaTqblAuKFBDO 7cjlfruK9Oqw7k5wAZGO72

        spoty spotify playlist remove-tracks https://open.spotify.com/playlist/37i9dQZF1DX8z1UW9HQvSq https://open.spotify.com/track/00i9VF7sjSaTqblAuKFBDO

    """
    track_ids = list(track_ids)
    spoty.spotify_api.remove_tracks_from_playlist(playlist_id, track_ids)
    click.echo(f'Tracks removed from playlist {playlist_id}')


@playlist.command("remove-liked-tracks")
@click.argument("playlist_ids", nargs=-1)
def playlist_remove_liked_tracks(playlist_ids):
    r"""
    Read playlists and remove all liked tracks found from these playlists.

    PLAYLIST_IDS - list of playlist IDs or URIs. You can specify one ID or many IDs separated by a space.

    Examples:

        spoty spotify playlist remove-liked-tracks 37i9dQZF1DX8z1UW9HQvSq

        spoty spotify playlist remove-liked-tracks 37i9dQZF1DX8z1UW9HQvSq 37i9dQZF1DX7jNFrjYQurt

        spoty spotify playlist remove-liked-tracks https://open.spotify.com/playlist/37i9dQZF1DX8z1UW9HQvSq

    """

    all_removed_tracks = []
    with click.progressbar(playlist_ids, label='Removing liked tracks from playlists') as bar:
        for playlist_id in bar:
            removed_tracks = spoty.spotify_api.remove_liked_tracks_in_playlist(playlist_id)
            all_removed_tracks.extend(removed_tracks)

    click.echo(f'{len(all_removed_tracks)} liked tracks removed from {len(playlist_ids)} playlists.')


@playlist.command("list-invalid-tracks")
@click.argument("playlist_ids", nargs=-1)
def playlist_list_invalid_tracks(playlist_ids):
    r"""
    Read playlists and list all invalid tracks found from these playlists.
    Tracks that have no ID are considered invalid.
    If a track is not available in the given region it is not considered invalid. You can still pull information on it. Tracks that are not available for the region will not be deleted.
    Invalid tracks are those for which there is no information at all in the database (have been deleted from the spotify database).

    PLAYLIST_IDS - list of playlist IDs or URIs. You can specify one ID or many IDs separated by a space.

    Examples:

        spoty spotify playlist list-invalid-tracks 37i9dQZF1DX8z1UW9HQvSq

        spoty spotify playlist list-invalid-tracks 37i9dQZF1DX8z1UW9HQvSq 37i9dQZF1DX7jNFrjYQurt

        spoty spotify playlist list-invalid-tracks https://open.spotify.com/playlist/37i9dQZF1DX8z1UW9HQvSq

    """

    all_invalid_tracks = []
    with click.progressbar(playlist_ids, label='Collecting invalid tracks from playlists') as bar:
        for playlist_id in bar:
            removed_tracks = spoty.spotify_api.get_invalid_tracks_in_playlist(playlist_id)
            all_invalid_tracks.extend(removed_tracks)

    click.echo(f'{len(all_invalid_tracks)} invalid tracks in {len(playlist_ids)} playlists.')

    if len(all_invalid_tracks) > 0:
        for track in all_invalid_tracks:
            click.echo(str(track))


@playlist.command("like-all-tracks")
@click.argument("playlist_ids", nargs=-1)
def playlist_like_all_tracks(playlist_ids):
    r"""
    Read playlists and like all tracks in these playlists.

    PLAYLIST_IDS - list of playlist IDs or URIs. You can specify one ID or many IDs separated by a space.

    Examples:

        spoty spotify playlist like-all-tracks 37i9dQZF1DX8z1UW9HQvSq

        spoty spotify playlist like-all-tracks 37i9dQZF1DX8z1UW9HQvSq 37i9dQZF1DX7jNFrjYQurt

        spoty spotify playlist like-all-tracks https://open.spotify.com/playlist/37i9dQZF1DX8z1UW9HQvSq

    """

    all_liked_tracks = []
    with click.progressbar(playlist_ids, label='Liking all tracks in playlists') as bar:
        for playlist_id in bar:
            liked_tracks = spoty.spotify_api.like_all_tracks_in_playlist(playlist_id)
            all_liked_tracks.extend(liked_tracks)

    click.echo(f'{len(all_liked_tracks)} tracks added to liked tracks in {len(playlist_ids)} playlists.')



import deezer.errors
from spoty import settings, secrets_file_name
from deezer import Deezer
from spoty import log
import spoty.utils
import os.path
import click
import time
import re
import datetime

DEEZER_TRACK_ID_TAG = 'DEEZER_TRACK_ID'

DEEZER_APP_ID = settings.default.DEEZER_APP_ID
DEEZER_APP_SECRET = settings.default.DEEZER_APP_SECRET
DEEZER_ACCESS_TOKEN = settings.default.DEEZER_ACCESS_TOKEN
DEEZER_REDIRECT_URI = settings.DEEZER.REDIRECT_URI
arl_file_name = os.path.join(spoty.config_path, '.arl')

dz = None


def get_dz():
    if DEEZER_APP_ID is None or DEEZER_APP_ID == "" or DEEZER_APP_ID == 0:
        click.echo(f"DEEZER_APP_ID is not set. Please, edit {secrets_file_name}", err=True)
        exit()

    if DEEZER_APP_SECRET is None or DEEZER_APP_SECRET == "":
        click.echo(f"DEEZER_APP_SECRET is not set. Please, edit {secrets_file_name}", err=True)
        exit()

    global dz
    if dz is None:
        dz = Deezer()
        get_arl()
    return dz


def request_valid_arl():
    while True:
        arl = input("Paste here your arl:")
        if get_dz().login_via_arl(arl.strip()): break
    return arl


def get_arl():
    if os.path.isfile(arl_file_name):
        with open(arl_file_name, 'r') as f:
            arl = f.readline().rstrip("\n").strip()
        if not get_dz().login_via_arl(arl): arl = request_valid_arl()
    else:
        arl = request_valid_arl()
    with open(arl_file_name, 'w') as f:
        f.write(arl)


def get_tracks_from_playlists(playlist_ids: list, add_spoty_tags=True):
    all_tracks = []
    all_tags_list = []
    all_received_playlists = []

    requested_playlists = []

    with click.progressbar(playlist_ids, label=f'Reading tracks in {len(playlist_ids)} Deezer playlists') as bar:
        for playlist_id in bar:

            # remove already requested playlists
            if playlist_id in requested_playlists:
                click.echo(f'Deezer playlist {playlist_id} requested twice. In will be skipped.')
                continue

            tracks, playlist = get_playlist_with_full_list_of_tracks(playlist_id, add_spoty_tags)
            requested_playlists.append(playlist_id)

            tags = read_tags_from_deezer_tracks(tracks)

            all_tracks.extend(tracks)
            all_tags_list.extend(tags)
            all_received_playlists.append(playlist_id)

    return all_tracks, all_tags_list, all_received_playlists


def get_tracks_of_deezer_user(user_id: str, playlists_names_regex: str = None):
    if user_id == 'me':
        playlists = get_list_of_user_playlists()
        click.echo(f'You have {len(playlists)} playlists in Deezer library')
    else:
        playlists = get_list_of_user_playlists(user_id)
        click.echo(f'User {user_id} has {len(playlists)} playlists in Deezer library')

    if playlists_names_regex is not None:
        playlists = list(filter(lambda pl: re.findall(playlists_names_regex, pl['title']), playlists))

    if len(playlists) == 0:
        return [], [], []

    ids = get_playlists_ids(playlists)
    tracks, tags, playlists = get_tracks_from_playlists(ids)

    return tracks, tags, playlists


def get_track_artist_and_title(track: dict):
    if track is None:
        return None

    artist = "Unknown"
    title = "Unknown"
    try:
        artist = track["ART_NAME"]
    except:
        pass

    try:
        title = track["SNG_TITLE"]
    except:
        pass

    return f"{artist} - {title}"


def get_track_ids(tracks: list):
    if len(tracks) == 0:
        return []
    else:
        return [str(track['SNG_ID']) for track in tracks]


def get_track_ids_from_tags_list(tags_list):
    if len(tags_list) == 0:
        return []
    else:
        return [str(track[DEEZER_TRACK_ID_TAG]) for track in tags_list]


def get_playlists_ids(playlists: list):
    if len(playlists) == 0:
        return []
    else:
        return [item['id'] for item in playlists]


def parse_playlist_id(id_or_url: str):
    id_or_url = str(id_or_url)
    if id_or_url.startswith("https://www.deezer.com/") and "/playlist/" in id_or_url:
        id_or_url = id_or_url.split('/playlist/')[1]
        id_or_url = id_or_url.split('?')[0]
    return id_or_url


def parse_track_id(id_or_url: str):
    id_or_url = str(id_or_url)
    if id_or_url.startswith("https://www.deezer.com/") and "/track/" in id_or_url:
        id_or_url = id_or_url.split('/track/')[1]
        id_or_url = id_or_url.split('?')[0]
    return id_or_url


def parse_user_id(id_or_url: str):
    id_or_url = str(id_or_url)
    if id_or_url.startswith("https://www.deezer.com/") and "/profile/" in id_or_url:
        id_or_url = id_or_url.split('/profile/')[1]
        id_or_url = id_or_url.split('?')[0]
    return id_or_url


def create_playlist(name: str):
    log.info(f'Creating new playlist')

    # "Quota exceeded for playlist_create" message is very often received.
    # todo: just send create playlist request with official API

    WAITING_DELAY = 60

    while True:
        try:
            playlist_id = get_dz().gw.create_playlist(name)
            return playlist_id
        except deezer.errors.GWAPIError as e:
            if e.args[0] == '{"QUOTA_ERROR": "Quota exceeded for playlist_create"}':
                click.echo(f"Quota exceeded for playlist_create. Waiting {WAITING_DELAY} seconds ...")
                time.sleep(WAITING_DELAY)


def get_playlist(playlist_id: str):
    playlist_id = parse_playlist_id(playlist_id)

    log.info(f'Requesting playlist {playlist_id}')

    playlist = get_dz().gw.get_playlist_page(playlist_id)
    return playlist['DATA']


def get_playlist_with_full_list_of_tracks(playlist_id: str, add_spoty_tags=True):
    playlist_id = parse_playlist_id(playlist_id)

    log.info(f'Collecting playlist {playlist_id}')

    playlist = get_dz().gw.get_playlist_page(playlist_id)
    tracks = get_dz().gw.get_playlist_tracks(playlist_id)

    playlist = playlist['DATA']

    log.debug(f'Playlist have {len(tracks)} tracks (playlist name: "{playlist["TITLE"]}")')

    if add_spoty_tags:
        add_spoty_tags_to_tracks([], tracks, playlist_id, playlist['TITLE'])

    return tracks, playlist


def get_playlists_with_full_list_of_tracks(playlist_ids: list):
    for i in range(len(playlist_ids)):
        playlist_ids[i] = parse_playlist_id(playlist_ids[i])

    playlists_dict = {}
    with click.progressbar(playlist_ids, label='Reading playlists') as bar:
        for playlist_id in bar:
            playlists_dict[playlist_id] = []

            tracks = get_dz().gw.get_playlist_tracks(playlist_id)
            playlists_dict[playlist_id].extend(tracks)

    return playlists_dict


def get_list_of_user_playlists(user_id: str = None):
    if user_id is None:
        user_id = parse_user_id(get_dz().current_user['id'])
        # user_data = get_dz().gw.get_user_data()
        # user_id = user_data['USER']['USER_ID']
    playlists = get_dz().gw.get_user_playlists(user_id)
    return playlists


def delete_all_playlists(confirm=False):
    log.info(f'Deleting all playlist')

    playlists = get_list_of_user_playlists()

    if len(playlists) <= 1:
        log.debug("User has no playlists.")
        return []

    if not confirm:
        if not click.confirm(f'Are you sure you want to delete all {len(playlists)} playlists?'):
            log.info("Playlists not deleted. Canceled by user.")
            return []

    ids = [pl['id'] for pl in playlists]
    deleted_playlists = delete_playlists(ids, True)

    return deleted_playlists


def delete_playlists(playlist_ids: list, confirm=False):
    deleted_playlists = []
    with click.progressbar(playlist_ids, label=f'Deleting {len(playlist_ids)} playlists') as bar:
        for playlist_id in bar:
            res = delete_playlist(playlist_id, confirm)
            if res:
                deleted_playlists.append(playlist_id)

    return deleted_playlists


def delete_playlist(playlist_id: str, confirm=False):
    playlist_id = parse_playlist_id(playlist_id)

    log.info(f'Deleting playlist {playlist_id}')

    playlist = get_playlist(playlist_id)
    if playlist == None:
        log.info(f'Playlist {playlist_id} does not exist.')
        return False

    if not confirm:
        if not click.confirm(f'\nAre you sure you want to delete playlist "{playlist["TITLE"]}"?'):
            log.info("Playlist not deleted. Canceled by user.")
            return False
        click.echo()

    res = get_dz().gw.delete_playlist(playlist_id)
    if res:
        log.success(f'Playlist {playlist_id} deleted"')
        return True
    else:
        log.error(f'Playlist {playlist_id} not deleted"')
        return False


def add_track_release_dates(tracks: list):
    with click.progressbar(tracks, label=f'Reading albums') as bar:
        for track in bar:
            if 'DEEZER_ALBUM_ID' in track:
                track['YEAR'] = get_album_release_date(track['DEEZER_ALBUM_ID'])
            elif 'ALB_ID' in track:
                track['YEAR'] = get_album_release_date(track['ALB_ID'])
    return tracks


def get_album_release_date(album_id: str):
    album = get_dz().gw.get_album(album_id)
    if 'ORIGINAL_RELEASE_DATE' in album:
        return album['ORIGINAL_RELEASE_DATE']
    if 'PHYSICAL_RELEASE_DATE' in album:
        return album['PHYSICAL_RELEASE_DATE']
    if 'DIGITAL_RELEASE_DATE' in album:
        return album['DIGITAL_RELEASE_DATE']
    return None


def find_missing_track_ids(tags_list: list, ignore_duration=False):
    found = []
    not_found = []

    tracks_without_id = [tags for tags in tags_list if not DEEZER_TRACK_ID_TAG in tags]

    if len(tracks_without_id) == 0:
        return tags_list, []

    with click.progressbar(tags_list, length=len(tracks_without_id),
                           label=f'Identifying {len(tracks_without_id)} tracks') as bar:
        for tags in tags_list:
            if DEEZER_TRACK_ID_TAG in tags:
                tags['SPOTY_FOUND_BY'] = DEEZER_TRACK_ID_TAG
                found.append(tags)
                continue

            if "ISRC" in tags:
                id = find_track_id_by_isrc(tags['ISRC'])
                if id is not None:
                    tags[DEEZER_TRACK_ID_TAG] = id
                    tags['SPOTY_FOUND_BY'] = 'ISRC'
                    found.append(tags)
                    bar.update(1)
                    continue

            if "TITLE" in tags and "ARTIST" in tags:
                if not ignore_duration and 'SPOTY_LENGTH' in tags:
                    id = find_track_id_by_artist_and_title(tags['ARTIST'], tags['TITLE'], tags['SPOTY_LENGTH'])
                else:
                    id = find_track_id_by_artist_and_title(tags['ARTIST'], tags['TITLE'])
                if id is not None:
                    tags[DEEZER_TRACK_ID_TAG] = id
                    if not ignore_duration:
                        tags['SPOTY_FOUND_BY'] = 'TITLE,ARTIST,SPOTY_LENGTH'
                    else:
                        tags['SPOTY_FOUND_BY'] = 'TITLE,ARTIST'
                    found.append(tags)
                    bar.update(1)
                    continue

            not_found.append(tags)
            bar.update(1)

    return found, not_found


def find_track_id_by_isrc(isrc: str):
    track = find_track_by_isrc(isrc)
    return track['id'] if track is not None else None


def find_track_by_isrc(isrc: str):
    try:
        track = get_dz().api.get_track_by_ISRC(isrc)
        log.debug(f'Track found by ISRC: {isrc} (ID: {track["id"]} TITLE: "{get_track_artist_and_title(track)}")')
        if track is not None and int(track['id']) != 0:
            return track
    except deezer.errors.DataException as e:
        log.debug(f'Track not found by ISRC: {isrc}"')

    return None


def find_track_id_by_artist_and_title(artist: str, title: str, length=None,
                                      length_tolerance=settings.SPOTY.COMPARE_LENGTH_TOLERANCE_SEC):
    track = find_track_by_artist_and_title(artist, title, length, length_tolerance)
    return track['id'] if track is not None else None


def find_track_by_artist_and_title(artist: str, title: str, length=None,
                                   length_tolerance=settings.SPOTY.COMPARE_LENGTH_TOLERANCE_SEC):
    if length is not None and length != 0 and length != "0":
        artist=artist.split(';')[0]
        tracks = get_dz().api.advanced_search(artist, "", title, "", int(length) - length_tolerance,
                                              int(length) + length_tolerance)
        if len(tracks['data']) > 0:
            return tracks['data'][0]
    else:
        tracks = get_dz().api.advanced_search(artist, "", title)
        if len(tracks['data']) > 0:
            return tracks['data'][0]
    return None


def find_track_by_id(id):
    track = get_dz().api.get_track(id)
    return track

def add_tracks_to_playlist_by_tags(playlist_id: str, tags_list: list, allow_duplicates=False):
    playlist_id = parse_playlist_id(playlist_id)

    tags_list = tags_list.copy()

    for i in range(len(tags_list)):
        tags_list[i][DEEZER_TRACK_ID_TAG] = parse_track_id(tags_list[i][DEEZER_TRACK_ID_TAG])

    import_duplicates = []

    if not allow_duplicates:
        tags_list, import_duplicates = spoty.utils.remove_duplicated_tags(tags_list, [DEEZER_TRACK_ID_TAG])
        if len(import_duplicates) > 0:
            log.debug(f'{len(import_duplicates)} duplicates found when adding tracks. It will be skipped.')

    log.info(f'Adding {len(tags_list)} tracks to playlist {playlist_id}')

    playlist = get_playlist(playlist_id)

    already_exist = []

    if not allow_duplicates:
        tracks, playlist = get_playlist_with_full_list_of_tracks(playlist_id)
        tags_in_playlist = read_tags_from_deezer_tracks(tracks)
        tags_list, already_exist = spoty.utils.remove_exist_tags(tags_in_playlist, tags_list, ['DEEZER_TRACK_ID'])
        if len(already_exist) > 0:
            log.debug(f'{len(already_exist)} tracks already exist and will be skipped.')

    track_ids = get_track_ids_from_tags_list(tags_list)

    # split by 100 tracks for one request
    tracks_count = 0
    playlist_index = 1
    while len(track_ids) > 0:
        tracks_to_proceed = track_ids[0:50]
        del track_ids[:50]

        get_dz().gw.add_songs_to_playlist(playlist_id, tracks_to_proceed)

        # split by 2000 tracks for one playlist
        tracks_count += len(tracks_to_proceed)
        if tracks_count >= 2000 and len(track_ids) > 0:
            tracks_count = 0
            playlist_index += 1
            title = f'{playlist["title"]} {playlist_index}'
            playlist = create_playlist(title)
            print(f'New playlist created. ID: {playlist["id"]} TITLE: {title}')

    log.success(f'Adding tracks complete (tracks added: {len(tags_list)}')

    return tags_list, import_duplicates, already_exist


def add_tracks_to_playlist_by_ids(playlist_id: str, track_ids: list, allow_duplicates=False):
    playlist_id = parse_playlist_id(playlist_id)

    import_duplicates = []

    for i in range(len(track_ids)):
        track_ids[i] = parse_track_id(track_ids[i])

    if not allow_duplicates:
        tags_list, import_duplicates = spoty.utils.remove_duplicates(track_ids)
        if len(import_duplicates) > 0:
            log.debug(f'{len(import_duplicates)} duplicates found when adding tracks. It will be skipped.')

    log.info(f'Adding {len(track_ids)} tracks to playlist {playlist_id}')

    playlist = get_playlist(playlist_id)

    already_exist = []

    if not allow_duplicates:
        tracks, playlist = get_playlist_with_full_list_of_tracks(playlist_id)
        ids_in_playlist = get_track_ids(tracks)
        track_ids, already_exist = spoty.utils.remove_exist(ids_in_playlist, track_ids)
        if len(already_exist) > 0:
            log.debug(f'{len(already_exist)} tracks already exist and will be skipped.')

    tracks_added = []
    tracks_copy = track_ids.copy()

    # split by 100 tracks for one request
    tracks_count = 0
    playlist_index = 1
    while len(tracks_copy) > 0:
        tracks_to_proceed = tracks_copy[0:50]
        del tracks_copy[:50]

        get_dz().gw.add_songs_to_playlist(playlist_id, tracks_to_proceed)
        tracks_added.extend(tracks_to_proceed)

        # split by 2000 tracks for one playlist
        tracks_count += len(tracks_to_proceed)
        if tracks_count >= 2000 and len(tracks_copy) > 0:
            tracks_count = 0
            playlist_index += 1
            title = f'{playlist["title"]} {playlist_index}'
            playlist = create_playlist(title)
            print(f'New playlist created. ID: {playlist["id"]} TITLE: {title}')

    log.success(f'Adding tracks complete (tracks added: {len(tracks_added)}')

    return tracks_added, import_duplicates, already_exist


def find_playlist_by_name(name):
    playlists = get_list_of_user_playlists()
    return list(filter(lambda pl: pl['title'] == name, playlists))


def import_playlists_from_tags_list(tags_list: list, grouping_pattern: str, overwrite_if_exist=False,
                                    append_if_exist=False,
                                    allow_duplicates=False, confirm=False):
    all_playlist_ids = []
    all_added = []
    all_source_duplicates = []
    all_already_exist = []
    grouped_tags = spoty.utils.group_tags_by_pattern(tags_list, grouping_pattern)

    with click.progressbar(grouped_tags.items(), label=f'Importing {len(grouped_tags)} playlists') as bar:
        for group_name, g_tags_list in bar:
            playlist_id, added, source_duplicates, already_exist \
                = import_playlist_from_tags_list(group_name, g_tags_list, overwrite_if_exist, append_if_exist,
                                                 allow_duplicates, confirm)

            all_playlist_ids.append(playlist_id)
            all_added.extend(added)
            all_source_duplicates.extend(source_duplicates)
            all_already_exist.extend(already_exist)

    return all_playlist_ids, all_added, all_source_duplicates, all_already_exist


def import_playlist_from_tags_list(playlist_name: str, tags_list: list, overwrite_if_exist=False, append_if_exist=False,
                                   allow_duplicates=False, confirm=False):
    if len(tags_list) == 0:
        return [], [], [], []

    log.info(f'Importing playlist "{playlist_name}"')

    playlist_id = None

    if overwrite_if_exist or append_if_exist:
        found_playlists = find_playlist_by_name(playlist_name)

        if len(found_playlists) > 0:
            if append_if_exist:
                if len(found_playlists) > 1:
                    if confirm:
                        click.echo(
                            f'\n{len(found_playlists)} playlists with name "{playlist_name}" found in Deezer library. Choosing the first one and appending it.')
                    else:
                        if not click.confirm(
                                f'\n{len(found_playlists)} playlists with name "{playlist_name}" found in Deezer library. Choose the first one and append it?'):
                            click.echo("\nCanceled")
                            log.info(f'Canceled by user (more than one playlists found with name "{playlist_name})"')
                            return [], [], [], []
                        click.echo()  # for new line
            if overwrite_if_exist:
                if len(found_playlists) > 1:
                    if confirm:
                        click.echo(
                            f'\n{len(found_playlists)} playlists with name "{playlist_name}" found in Deezer library. Choosing the first one and overwriting it.')
                    else:
                        if not click.confirm(
                                f'\n{len(found_playlists)} playlists with name "{playlist_name}" found in Deezer library. Choose the first one and overwrite it?'):
                            click.echo("\nCanceled")
                            log.info(f'Canceled by user (more than one playlists found with name "{playlist_name})"')
                            return [], [], [], []
                        click.echo()  # for new line
                else:
                    if confirm:
                        click.echo(f'\nPlaylist "{playlist_name}" exist in Deezer library. Overwriting it.')
                    else:
                        if not click.confirm(
                                f'\nPlaylist "{playlist_name}" exist in Deezer library. Overwrite it?'):
                            click.echo("\nCanceled")
                            log.info(f'Canceled by user (playlist found with name "{playlist_name})"')
                            return [], [], [], []
                        click.echo()  # for new line

                remove_all_tracks_from_playlist(found_playlists[0]['id'], True)

            playlist_id = found_playlists[0]['id']

    if playlist_id is None:
        playlist_id = create_playlist(playlist_name)

    added, source_duplicates, already_exist = add_tracks_to_playlist_by_tags(playlist_id, tags_list, allow_duplicates)

    log.success(f'Playlist imported (new tracks: "{len(added)}")  id: {playlist_id} name: "{playlist_name}"')

    return playlist_id, added, source_duplicates, already_exist


def remove_all_tracks_from_playlist(playlist_id: str, confirm=False):
    playlist_id = parse_playlist_id(playlist_id)

    playlist = get_playlist_with_full_list_of_tracks(playlist_id)
    tracks = playlist.tracks

    if len(tracks) == 0:
        return False

    if not confirm:
        if not click.confirm(f'Do you want to remove all tracks from playlist {playlist_id}?'):
            click.echo("\nCanceled")
            return False
        click.echo()  # for new line

    ids = get_track_ids(tracks)
    remove_tracks_from_playlist(playlist_id, ids)
    return True


def remove_tracks_from_playlist(playlist_id: str, track_ids: list):
    playlist_id = parse_playlist_id(playlist_id)

    log.info(f'Removing {len(track_ids)} tracks from playlist {playlist_id}')

    with click.progressbar(track_ids, label=f'Deleting {len(track_ids)} tracks') as bar:
        for track_id in bar:
            get_dz().gw.remove_song_from_playlist(playlist_id, track_id)

    # get_dz().gw.remove_songs_from_playlist(playlist_id, track_ids)

    # split request by 100 songs
    # i = 0
    # next_tracks = []
    # while i < len(track_ids):
    #     track_ids[i] = parse_track_id(track_ids[i])
    #     next_tracks.append(track_ids[i])
    #     if len(next_tracks) == 100 or i == len(track_ids) - 1:
    #         get_dz().gw.remove_songs_from_playlist(playlist_id, next_tracks)
    #         next_tracks = []
    #     i += 1

    log.success(f'Tracks removed from playlist {playlist_id}')


def add_spoty_tags_to_tracks(tracks: list, new_tracks: list, playlist_id: str, playlist_name: str):
    counter = len(tracks)
    for track in new_tracks:
        track['SPOTY_PLAYLIST_ID'] = playlist_id
        track['SPOTY_PLAYLIST_INDEX'] = str(counter + 1)
        track['SPOTY_SOURCE'] = 'DEEZER'
        if playlist_name is not None:
            track['SPOTY_PLAYLIST_NAME'] = playlist_name
        try:
            track['SPOTY_TRACK_ID'] = track['SNG_ID']
        except:
            pass
        counter += 1
    return counter


def read_tags_from_deezer_tracks(tracks: list):
    tag_tracks = []

    for track in tracks:
        tags = read_tags_from_deezer_track(track)
        tag_tracks.append(tags)

    return tag_tracks


def read_tags_from_deezer_track(track: dict):
    tags = {}

    if 'isrc' in track:
        tags['ISRC'] = track['isrc']


    if 'artist' in track and 'name' in track['artist']:
        tags['ARTIST'] = track['artist']['name']


    if 'title' in track:
        tags['TITLE'] = track['title']



    if 'duration' in track:
        tags['SPOTY_LENGTH'] = int(track['duration'])

    if 'album' in track:
        tags['DEEZER_ALBUM_ID'] = track['album']['id']

    if 'id' in track:
        tags['WWWAUDIOFILE'] = f'https://www.deezer.com/track/{track["id"]}'

    tags['DEEZER_TRACK_ID'] = track["id"]

    if 'artist' in track and 'id' in track['artist']:
        tags['DEEZER_ARTIST_ID'] = track['artist']['id']

    if 'explicit_lyrics' in track:
        tags['EXPLICIT'] = track['explicit_lyrics']



    if 'release_date' in track:
        tags['YEAR'] = track['release_date']

    if 'gain' in track:
        tags['GAIN'] = track['gain']

    if 'bpm' in track:
        tags['BPM'] = track['bpm']

    # if 'DATE_ADD' in track:
    #     timestamp = datetime.datetime.fromtimestamp(int(track['DATE_ADD']))
    #     tags['DEZZER_DATE_ADD'] = track['DATE_ADD']
        # tags['SPOTY_TRACK_ADDED'] = timestamp.strftime('%Y-%m-%d %H:%M:%S')


    for tag in spoty.utils.spoty_tags:
        if tag in track:
            tags[tag] = track[tag]

    tags = spoty.utils.clean_tags_after_read(tags)

    return tags




from spoty import settings
from spoty import log
from spoty.utils import SpotyContext
import spoty.utils
import click


@click.command("count")
@click.option('--grouping-pattern', '--gp', show_default=True,
              default=settings.SPOTY.DEFAULT_GROUPING_PATTERN,
              help='Tracks will be grouped to playlists according to this pattern.')
@click.pass_obj
def count_tracks(context: SpotyContext,
                 grouping_pattern,
                 ):
    """
Print a number of tracks to console.
    """
    context.summary.append(f'Counting:')

    for i, tags_list in enumerate(context.tags_lists):
        grouped_tags = spoty.utils.group_tags_by_pattern(tags_list, grouping_pattern)

        if len(context.tags_lists) > 1:
            context.summary.append(f'  --- List {i + 1}:')

        context.summary.append(f'  {len(tags_list)} tracks.')

        if len(grouped_tags) > 0:
            context.summary.append(f'  {len(grouped_tags)} groups by pattern.')

    if len(context.duplicates_groups) > 0:
        context.summary.append(f'  ---')
        all_def_list = []
        all_prob_list = []
        all_source_list = []
        for group in context.duplicates_groups:
            all_source_list.append(group.source_tags)
            all_def_list.extend(group.def_duplicates)
            all_prob_list.extend(group.prob_duplicates)

        context.summary.append(f'  {len(all_source_list)} duplicate groups.')
        context.summary.append(f'  {len(all_def_list)} definitely duplicates.')
        context.summary.append(f'  {len(all_prob_list)} probably duplicates.')

    if len(context.unique_first_tracks) > 0 or len(context.unique_second_tracks) > 0:
        context.summary.append(f'  ---')
        context.summary.append(f'  {len(context.unique_first_tracks)} unique tracks in fist list.')
        context.summary.append(f'  {len(context.unique_second_tracks)} unique tracks in second list.')

    click.echo('\n------------------------------------------------------------')
    click.echo('\n'.join(context.summary))



from spoty import log
import spoty.utils
import os.path
import click
import time, datetime
from mutagen.flac import FLAC
from mutagen.mp3 import MP3
from mutagen.easyid3 import EasyID3
from mutagen.id3 import ID3, TXXX, TMOO


def is_flac(file_name):
    return file_name.upper().endswith('.FLAC')


def is_mp3(file_name):
    return file_name.upper().endswith('.MP3')


def is_audio_file(file_name):
    return is_flac(file_name) or is_mp3(file_name)


def find_audio_files_in_paths(paths, recursive=True):
    all_file_names = []

    for path in paths:
        file_names = find_audio_files_in_path(path, recursive)
        all_file_names.extend(file_names)

    return all_file_names


def find_audio_files_in_path(path, recursive=True):
    path = os.path.abspath(path)

    if not spoty.utils.is_valid_path(path):
        raise Exception("Path is not valid: " + path)

    file_names = []

    if recursive:
        file_names = [os.path.join(dp, f) for dp, dn, filenames in os.walk(path) for f in filenames if
                      is_flac(os.path.splitext(f)[1]) or is_mp3(os.path.splitext(f)[1])]
    else:
        file_names = [os.path.join(path, f) for f in os.listdir(path) if
                      os.path.isfile(os.path.join(path, f))]
        file_names = list(
            filter(lambda f: is_flac(f) or is_mp3(f), file_names))

    return file_names


def write_audio_file_tags(file_name, new_tags):
    if len(new_tags) > 0:
        if is_flac(file_name):
            f = FLAC(file_name)
            for key, value in new_tags.items():
                f[key] = str(value)
            f.save()
        if is_mp3(file_name):
            f = EasyID3(file_name)
            for key, value in new_tags.items():
                if key.lower() not in f.valid_keys:
                    f.RegisterTXXXKey(key.lower(), key.upper())
                f[key.lower()] = str(value)
            f.save(v2_version=3)
            f = ID3(file_name)
            edited = False
            for key, value in new_tags.items():
                if key == "MOOD":
                    edited = True
                    f.add(TMOO(encoding=3, text=str(value)))
            if edited:
                f.save(v2_version=3)


def read_audio_files_tags(file_names, add_spoty_tags=True, clean_tags=True):
    tags_list = []
    with click.progressbar(file_names, label=f'Reading tags in {len(file_names)} files') as bar:
        for file_name in bar:
            tags = read_audio_file_tags(file_name, add_spoty_tags, clean_tags)
            if tags is not None:
                tags_list.append(tags)
    return tags_list


def read_audio_file_tags(file_name, add_spoty_tags=True, clean_tags=True):
    tags = {}

    file_name = os.path.abspath(file_name)

    if not spoty.utils.is_valid_file(file_name):
        click.echo(f"\nFile not found: {file_name}")
        return None

    if add_spoty_tags:
        dir = os.path.dirname(file_name)
        tags['SPOTY_FILE_NAME'] = file_name
        tags['SPOTY_SOURCE'] = "LOCAL"
        tags['SPOTY_PLAYLIST_NAME'] = os.path.basename(os.path.normpath(dir))
        tags['SPOTY_TRACK_ADDED'] = datetime.datetime.fromtimestamp(os.path.getctime(file_name)).strftime(
            '%Y-%m-%d %H:%M:%S')

    if is_flac(file_name):
        try:
            f = FLAC(file_name)
            tags['SPOTY_LENGTH'] = str(int(f.info.length))
            for tag in f.tags:
                if tag[0] in tags:  # adding same key with one more value
                    tags[tag[0]] += ';' + tag[1]
                else:
                    tags[tag[0]] = tag[1]
        except:
            click.echo(f"\nCant open file: {file_name}")
            return None

    if is_mp3(file_name):
        try:
            f = MP3(file_name, ID3=EasyID3)
            tags['SPOTY_LENGTH'] = str(int(f.info.length))
            f = EasyID3(file_name)
            keys = f.keys()
            for tag in f.valid_keys.keys():
                if tag in f:
                    tag_val = ';'.join(f[tag])
                    tags[tag.upper()] = tag_val
            f = ID3(file_name)
            for txxx in f.getall("TXXX"):  # custom keys
                tag = txxx.desc.upper()
                val = ';'.join(txxx.text)
                tags[tag] = val
        except:
            click.echo(f"\nCant read file: {file_name}")
            return None

    if clean_tags:
        tags = spoty.utils.clean_tags_after_read(tags)

    return tags
